# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.python import failure
from twisted.python import log

from buildbot.util import Notifier


class DeferWaiter:
    """This class manages a set of Deferred objects and allows waiting for their completion"""

    def __init__(self):
        self._waited_count = 0
        self._finish_notifier = Notifier()

    def _finished(self, result, d):
        # most likely nothing is consuming the errors, so do it here
        if isinstance(result, failure.Failure):
            log.err(result)

        self._waited_count -= 1
        if self._waited_count == 0:
            self._finish_notifier.notify(None)
        return result

    def add(self, d):
        if not isinstance(d, defer.Deferred):
            return None

        self._waited_count += 1
        d.addBoth(self._finished, d)
        return d

    def has_waited(self):
        return self._waited_count > 0

    @defer.inlineCallbacks
    def wait(self):
        if self._waited_count == 0:
            return
        yield self._finish_notifier.wait()


class RepeatedActionHandler:
    """This class handles a repeated action such as submitting keepalive requests. It integrates
    with DeferWaiter to correctly control shutdown of such process.
    """

    def __init__(self, reactor, waiter, interval, action, start_timer_after_action_completes=False):
        self._reactor = reactor
        self._waiter = waiter
        self._interval = interval
        self._action = action
        self._enabled = False
        self._timer = None
        self._start_timer_after_action_completes = start_timer_after_action_completes
        self._running = False

    def set_interval(self, interval):
        self._interval = interval

    def start(self):
        if self._enabled:
            return
        self._enabled = True
        self._start_timer()

    def stop(self):
        if not self._enabled:
            return

        self._enabled = False
        if self._timer:
            self._timer.cancel()
            self._timer = None

    def delay(self):
        if not self._enabled or not self._timer:
            # If self._timer is None, then the action is running and timer will be started once
            # it's done.
            return
        self._timer.reset(self._interval)

    def force(self):
        if not self._enabled or self._running:
            return

        self._timer.cancel()
        self._waiter.add(self._handle_action())

    def _start_timer(self):
        self._timer = self._reactor.callLater(self._interval, self._handle_timeout)

    @defer.inlineCallbacks
    def _do_action(self):
        try:
            self._running = True
            yield self._action()
        except Exception as e:
            log.err(e, 'Got exception in RepeatedActionHandler')
        finally:
            self._running = False

    def _handle_timeout(self):
        self._waiter.add(self._handle_action())

    @defer.inlineCallbacks
    def _handle_action(self):
        self._timer = None
        if self._start_timer_after_action_completes:
            yield self._do_action()

        if self._enabled:
            self._start_timer()

        if not self._start_timer_after_action_completes:
            yield self._do_action()


class NonRepeatedActionHandler:
    """This class handles a single action that can be issued on demand. It ensures that multiple
    invocations of an action do not overlap.
    """

    def __init__(self, reactor, waiter, action):
        self._reactor = reactor
        self._waiter = waiter
        self._action = action
        self._timer = None
        self._running = False
        self._repeat_after_finished = False

    def force(self, invoke_again_if_running=False):
        if self._running:
            if not invoke_again_if_running:
                return
            self._repeat_after_finished = True
            return

        if self._timer is not None:
            self._timer.cancel()
            self._timer = None

        self._waiter.add(self._do_action())

    def schedule(self, seconds_from_now, invoke_again_if_running=False):
        if self._running and not invoke_again_if_running:
            return

        if self._timer is None:
            self._timer = self._reactor.callLater(seconds_from_now, self._handle_timeout)
            return

        target_time = self._reactor.seconds() + seconds_from_now
        if target_time > self._timer.getTime():
            return

        self._timer.reset(seconds_from_now)

    def stop(self):
        if self._timer:
            self._timer.cancel()
            self._timer = None

    @defer.inlineCallbacks
    def _do_action(self):
        try:
            self._running = True
            yield self._action()
        except Exception as e:
            log.err(e, 'Got exception in NonRepeatedActionHandler')
        finally:
            self._running = False
        if self._repeat_after_finished:
            self._repeat_after_finished = False
            self._waiter.add(self._do_action())

    def _handle_timeout(self):
        self._timer = None
        self._waiter.add(self._do_action())



# This file is part of Buildbot. Buildbot is free software: you can)
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json as jsonmodule

from twisted.internet import defer
from twisted.logger import Logger
from twisted.python import deprecate
from twisted.python import versions
from twisted.python.threadpool import ThreadPool
from twisted.web.client import Agent
from twisted.web.client import HTTPConnectionPool
from zope.interface import implementer

from buildbot.interfaces import IHttpResponse
from buildbot.util import service
from buildbot.util import toJson
from buildbot.util import unicode2bytes

try:
    import txrequests
except ImportError:
    txrequests = None

import treq

log = Logger()


@implementer(IHttpResponse)
class TxRequestsResponseWrapper:
    def __init__(self, res):
        self._res = res

    def content(self):
        return defer.succeed(self._res.content)

    def json(self):
        return defer.succeed(self._res.json())

    @property
    def code(self):
        return self._res.status_code

    @property
    def url(self):
        return self._res.url


@implementer(IHttpResponse)
class TreqResponseWrapper:
    def __init__(self, res):
        self._res = res

    def content(self):
        return self._res.content()

    def json(self):
        return self._res.json()

    @property
    def code(self):
        return self._res.code

    @property
    def url(self):
        return self._res.request.absoluteURI.decode()


class HTTPClientService(service.SharedService):
    """A SharedService class that can make http requests to remote services.

    I provide minimal get/post/put/delete API with automatic baseurl joining, and json data encoding
    that is suitable for use from buildbot services.
    """

    # Those could be in theory be overridden in master.cfg by using
    # import buildbot.util.httpclientservice.HTTPClientService.PREFER_TREQ = True
    # We prefer at the moment keeping it simple
    PREFER_TREQ = False
    MAX_THREADS = 20

    def __init__(
        self,
        base_url,
        auth=None,
        headers=None,
        verify=None,
        cert=None,
        debug=False,
        skipEncoding=False,
    ):
        super().__init__()
        self._session = HTTPSession(
            self,
            base_url,
            auth=auth,
            headers=headers,
            verify=verify,
            cert=cert,
            debug=debug,
            skip_encoding=skipEncoding,
        )
        self._pool = None
        self._txrequests_sessions = []

    def updateHeaders(self, headers):
        self._session.update_headers(headers)

    @staticmethod
    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0))
    def checkAvailable(from_module):
        pass

    def startService(self):
        if txrequests is not None:
            self._txrequests_pool = ThreadPool(minthreads=1, maxthreads=self.MAX_THREADS)
            # unclosed ThreadPool leads to reactor hangs at shutdown
            # this is a problem in many situation, so better enforce pool stop here
            self.master.reactor.addSystemEventTrigger(
                "after",
                "shutdown",
                lambda: self._txrequests_pool.stop() if self._txrequests_pool.started else None,
            )
            self._txrequests_pool.start()

        self._pool = HTTPConnectionPool(self.master.reactor)
        self._pool.maxPersistentPerHost = self.MAX_THREADS
        return super().startService()

    @defer.inlineCallbacks
    def stopService(self):
        if txrequests is not None:
            sessions = self._txrequests_sessions
            self._txrequests_sessions = []
            for session in sessions:
                session.close()
            self._txrequests_pool.stop()
        if self._pool:
            yield self._pool.closeCachedConnections()
        yield super().stopService()

    def _do_request(self, session, method, ep, **kwargs):
        prefer_treq = self.PREFER_TREQ
        if session.auth is not None and not isinstance(session.auth, tuple):
            prefer_treq = False
        if prefer_treq or txrequests is None:
            return self._do_treq(session, method, ep, **kwargs)
        else:
            return self._do_txrequest(session, method, ep, **kwargs)

    def _prepare_request(self, session, ep, kwargs):
        if ep.startswith('http://') or ep.startswith('https://'):
            url = ep
        else:
            assert ep == "" or ep.startswith("/"), "ep should start with /: " + ep
            url = session.base_url + ep
        if session.auth is not None and 'auth' not in kwargs:
            kwargs['auth'] = session.auth
        headers = kwargs.get('headers', {})
        if session.headers is not None:
            headers.update(session.headers)
        kwargs['headers'] = headers

        # we manually do the json encoding in order to automatically convert timestamps
        # for txrequests and treq
        json = kwargs.pop('json', None)
        if isinstance(json, (dict, list)):
            jsonStr = jsonmodule.dumps(json, default=toJson)
            kwargs['headers']['Content-Type'] = 'application/json'
            if session.skip_encoding:
                kwargs['data'] = jsonStr
            else:
                jsonBytes = unicode2bytes(jsonStr)
                kwargs['data'] = jsonBytes
        return url, kwargs

    @defer.inlineCallbacks
    def _do_txrequest(self, session, method, ep, **kwargs):
        url, kwargs = yield self._prepare_request(session, ep, kwargs)
        if session.debug:
            log.debug("http {url} {kwargs}", url=url, kwargs=kwargs)

        def readContent(txrequests_session, res):
            # this forces reading of the content inside the thread
            _ = res.content
            if session.debug:
                log.debug("==> {code}: {content}", code=res.status_code, content=res.content)
            return res

        # read the whole content in the thread
        kwargs['background_callback'] = readContent
        if session.verify is False:
            kwargs['verify'] = False
        elif session.verify:
            kwargs['verify'] = session.verify

        if session.cert:
            kwargs['cert'] = session.cert
        if session._txrequests_session is None:
            session._txrequests_session = txrequests.Session(
                pool=self._txrequests_pool, maxthreads=self.MAX_THREADS
            )
            # FIXME: remove items from the list as HTTPSession objects are destroyed
            self._txrequests_sessions.append(session._txrequests_session)

        res = yield session._txrequests_session.request(method, url, **kwargs)
        return IHttpResponse(TxRequestsResponseWrapper(res))

    @defer.inlineCallbacks
    def _do_treq(self, session, method, ep, **kwargs):
        url, kwargs = yield self._prepare_request(session, ep, kwargs)
        # treq requires header values to be an array
        if "headers" in kwargs:
            kwargs['headers'] = {k: [v] for k, v in kwargs["headers"].items()}

        if session._treq_agent is None:
            session._trex_agent = Agent(self.master.reactor, pool=self._pool)
        kwargs['agent'] = session._trex_agent

        res = yield getattr(treq, method)(url, **kwargs)
        return IHttpResponse(TreqResponseWrapper(res))

    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), "Use HTTPSession.get()")
    def get(self, ep, **kwargs):
        return self._do_request(self._session, 'get', ep, **kwargs)

    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), "Use HTTPSession.put()")
    def put(self, ep, **kwargs):
        return self._do_request(self._session, 'put', ep, **kwargs)

    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), "Use HTTPSession.delete()")
    def delete(self, ep, **kwargs):
        return self._do_request(self._session, 'delete', ep, **kwargs)

    @deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), "Use HTTPSession.post()")
    def post(self, ep, **kwargs):
        return self._do_request(self._session, 'post', ep, **kwargs)


class HTTPSession:
    def __init__(
        self,
        http,
        base_url,
        auth=None,
        headers=None,
        verify=None,
        cert=None,
        debug=False,
        skip_encoding=False,
    ):
        assert not base_url.endswith("/"), "baseurl should not end with /: " + base_url
        self.http = http
        self.base_url = base_url
        self.auth = auth
        self.headers = headers
        self.pool = None
        self.verify = verify
        self.cert = cert
        self.debug = debug
        self.skip_encoding = skip_encoding

        self._treq_agent = None
        self._txrequests_session = None

    def update_headers(self, headers):
        if self.headers is None:
            self.headers = {}
        self.headers.update(headers)

    def get(self, ep, **kwargs):
        return self.http._do_request(self, 'get', ep, **kwargs)

    def put(self, ep, **kwargs):
        return self.http._do_request(self, 'put', ep, **kwargs)

    def delete(self, ep, **kwargs):
        return self.http._do_request(self, 'delete', ep, **kwargs)

    def post(self, ep, **kwargs):
        return self.http._do_request(self, 'post', ep, **kwargs)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from contextlib import contextmanager

from sqlalchemy.exc import OperationalError
from sqlalchemy.exc import ProgrammingError

from buildbot.config.master import DBConfig as MasterDBConfig
from buildbot.config.master import MasterConfig
from buildbot.db import enginestrategy
from buildbot.db import model
from buildbot.db import state


class FakeDBConnector:
    def __init__(self, engine):
        self.pool = FakePool(engine)
        self.master = FakeMaster()
        self.model = model.Model(self)
        self.state = state.StateConnectorComponent(self)

    @contextmanager
    def connect(self):
        try:
            with self.pool.engine.connect() as conn:
                yield conn
        finally:
            self.pool.engine.dispose()


class FakeCacheManager:
    def get_cache(self, cache_name, miss_fn):
        return None


class FakeMaster:
    def __init__(self):
        self.caches = FakeCacheManager()


class FakePool:
    def __init__(self, engine):
        self.engine = engine


class DbConfig:
    db_config: MasterDBConfig

    def __init__(self, BuildmasterConfig, basedir, name="config"):
        self.db_config = MasterConfig.get_dbconfig_from_config(BuildmasterConfig, throwErrors=False)
        self.basedir = basedir
        self.name = name

    def getDb(self):
        try:
            db = FakeDBConnector(
                engine=enginestrategy.create_engine(self.db_config.db_url, basedir=self.basedir)
            )
        except Exception:
            # db_config.db_url is probably trash. Just ignore, config.py db part will
            # create proper message
            return None

        with db.connect() as conn:
            try:
                self.objectid = db.state.thdGetObjectId(conn, self.name, "DbConfig")['id']
            except (ProgrammingError, OperationalError):
                conn.rollback()
                # ProgrammingError: mysql&pg, OperationalError: sqlite
                # assume db is not initialized
                return None

        return db

    def get(self, name, default=state.StateConnectorComponent.Thunk):
        db = self.getDb()
        if db is not None:
            with db.connect() as conn:
                ret = db.state.thdGetState(conn, self.objectid, name, default=default)
        else:
            if default is not state.StateConnectorComponent.Thunk:
                return default
            raise KeyError("Db not yet initialized")
        return ret

    def set(self, name, value):
        db = self.getDb()
        if db is not None:
            with db.connect() as conn:
                db.state.thdSetState(conn, self.objectid, name, value)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


# NOTE regarding LDAP encodings:
#
# By default the encoding used in ldap3 is utf-8. The encoding is user-configurable, though.
# For more information check ldap3's documentation on this topic:
# http://ldap3.readthedocs.io/encoding.html
#
# It is recommended to use ldap3's auto-decoded `attributes` values for
# `unicode` and `raw_*` attributes for `bytes`.


from __future__ import annotations

import importlib
from typing import Any
from urllib.parse import urlparse

from twisted.internet import defer
from twisted.internet import threads

from buildbot.util import bytes2unicode
from buildbot.util import flatten
from buildbot.www import auth
from buildbot.www import avatar

try:
    import ldap3
except ImportError:
    ldap3 = None


class LdapUserInfo(avatar.AvatarBase, auth.UserInfoProviderBase):
    name = 'ldap'

    def __init__(
        self,
        uri: str,
        bindUser: str | None,
        bindPw: str | None,
        accountBase: str,
        accountPattern: str,
        accountFullName: str,
        accountEmail: str,
        groupBase: str | None = None,
        groupMemberPattern: str | None = None,
        groupName: str | None = None,
        avatarPattern: str | None = None,
        avatarData: str | None = None,
        accountExtraFields: list[str] | None = None,
        tls: object | None = None,
    ) -> None:
        # Throw import error now that this is being used
        if not ldap3:
            importlib.import_module('ldap3')
        self.uri = uri
        self.bindUser = bindUser
        self.bindPw = bindPw
        self.accountBase = accountBase
        self.accountEmail = accountEmail
        self.accountPattern = accountPattern
        self.accountFullName = accountFullName
        group_params = [p for p in (groupName, groupMemberPattern, groupBase) if p is not None]
        if len(group_params) not in (0, 3):
            raise ValueError(
                "Incomplete LDAP groups configuration. "
                "To use Ldap groups, you need to specify the three "
                "parameters (groupName, groupMemberPattern and groupBase). "
            )

        self.groupName = groupName
        self.groupMemberPattern = groupMemberPattern
        self.groupBase = groupBase
        self.avatarPattern = avatarPattern
        self.avatarData = avatarData
        if accountExtraFields is None:
            accountExtraFields = []
        self.accountExtraFields = accountExtraFields
        self.ldap_encoding = ldap3.get_config_parameter('DEFAULT_SERVER_ENCODING')
        self.tls = tls

    def connectLdap(self) -> ldap3.Connection:
        server = urlparse(self.uri)
        netloc = server.netloc.split(":")
        # define the server and the connection
        s = ldap3.Server(
            netloc[0],
            port=int(netloc[1]),
            use_ssl=server.scheme == 'ldaps',
            get_info=ldap3.ALL,
            tls=self.tls,
        )

        auth = ldap3.SIMPLE
        if self.bindUser is None and self.bindPw is None:
            auth = ldap3.ANONYMOUS

        c = ldap3.Connection(
            s,
            auto_bind=True,
            client_strategy=ldap3.SYNC,
            user=self.bindUser,
            password=self.bindPw,
            authentication=auth,
        )
        return c

    def search(
        self,
        c: ldap3.Connection,
        base: str | None,
        filterstr: str | None = 'f',
        attributes: list[str | None] | None = None,
    ) -> list:
        c.search(base, filterstr, ldap3.SUBTREE, attributes=attributes)
        return c.response

    def getUserInfo(self, username: str) -> defer.Deferred:
        username = bytes2unicode(username)

        def thd() -> dict[str, object]:
            c = self.connectLdap()
            infos: dict[str, Any] = {'username': username}
            pattern = self.accountPattern % {"username": username}
            res = self.search(
                c,
                self.accountBase,
                pattern,
                attributes=[self.accountEmail, self.accountFullName, *self.accountExtraFields],
            )
            if len(res) != 1:
                raise KeyError(f"ldap search \"{pattern}\" returned {len(res)} results")
            dn, ldap_infos = res[0]['dn'], res[0]['attributes']

            def getFirstLdapInfo(x: list | None) -> str | None:
                if isinstance(x, list):
                    x = x[0] if x else None
                return x

            infos['full_name'] = getFirstLdapInfo(ldap_infos[self.accountFullName])
            infos['email'] = getFirstLdapInfo(ldap_infos[self.accountEmail])
            for f in self.accountExtraFields:
                if f in ldap_infos:
                    infos[f] = getFirstLdapInfo(ldap_infos[f])

            if self.groupMemberPattern is None:
                infos['groups'] = []
                return infos

            # needs double quoting of backslashing
            pattern = self.groupMemberPattern % {"dn": ldap3.utils.conv.escape_filter_chars(dn)}
            res = self.search(c, self.groupBase, pattern, attributes=[self.groupName])
            infos['groups'] = flatten([
                group_infos['attributes'][self.groupName] for group_infos in res
            ])

            return infos

        return threads.deferToThread(thd)

    def findAvatarMime(self, data: bytes) -> tuple[bytes, bytes] | None:
        # http://en.wikipedia.org/wiki/List_of_file_signatures
        if data.startswith(b"\xff\xd8\xff"):
            return (b"image/jpeg", data)
        if data.startswith(b"\x89PNG"):
            return (b"image/png", data)
        if data.startswith(b"GIF8"):
            return (b"image/gif", data)
        # ignore unknown image format
        return None

    def getUserAvatar(
        self, email: bytes, username: bytes | None, size: int, defaultAvatarUrl: str
    ) -> defer.Deferred:
        username_str = bytes2unicode(username) if username is not None else None
        email_str = bytes2unicode(email) if email is not None else None

        def thd() -> tuple[bytes, bytes] | None:
            c = self.connectLdap()
            if username_str:
                pattern = self.accountPattern % {"username": username_str}
            elif email_str:
                pattern = self.avatarPattern % {"email": email_str}  # type: ignore[operator]
            else:
                return None
            res = self.search(c, self.accountBase, pattern, attributes=[self.avatarData])
            if not res:
                return None
            ldap_infos = res[0]['raw_attributes']
            if ldap_infos.get(self.avatarData):
                data = ldap_infos[self.avatarData][0]
                return self.findAvatarMime(data)
            return None

        return threads.deferToThread(thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer


@defer.inlineCallbacks
def get_real_locks_from_accesses_raw(locks, props, builder, workerforbuilder, config_version):
    workername = workerforbuilder.worker.workername

    if props is not None:
        locks = yield props.render(locks)

    if not locks:
        return []

    locks = yield builder.botmaster.getLockFromLockAccesses(locks, config_version)
    return [(l.getLockForWorker(workername), a) for l, a in locks]


def get_real_locks_from_accesses(locks, build):
    return get_real_locks_from_accesses_raw(
        locks, build, build.builder, build.workerforbuilder, build.config_version
    )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import enum

from twisted.internet import defer
from twisted.python import log
from zope.interface import implementer

from buildbot import interfaces
from buildbot.machine.base import Machine
from buildbot.util import Notifier


class States(enum.Enum):
    # Represents the state of LatentMachine
    STOPPED = 0
    STARTING = 1
    STARTED = 2
    STOPPING = 3


@implementer(interfaces.ILatentMachine)
class AbstractLatentMachine(Machine):
    DEFAULT_MISSING_TIMEOUT = 20 * 60

    def checkConfig(
        self, name, build_wait_timeout=0, missing_timeout=DEFAULT_MISSING_TIMEOUT, **kwargs
    ):
        super().checkConfig(name, **kwargs)
        self.state = States.STOPPED
        self.latent_workers = []

    @defer.inlineCallbacks
    def reconfigService(
        self, name, build_wait_timeout=0, missing_timeout=DEFAULT_MISSING_TIMEOUT, **kwargs
    ):
        yield super().reconfigService(name, **kwargs)
        self.build_wait_timeout = build_wait_timeout
        self.missing_timeout = missing_timeout

        for worker in self.workers:
            if not interfaces.ILatentWorker.providedBy(worker):
                raise RuntimeError(f'Worker is not latent {worker.name}')

        self.state = States.STOPPED
        self._start_notifier = Notifier()
        self._stop_notifier = Notifier()
        self._build_wait_timer = None
        self._missing_timer = None

    def start_machine(self):
        # Responsible for starting the machine. The function should return a
        # deferred which should result in True if the startup has been
        # successful, or False otherwise.
        raise NotImplementedError

    def stop_machine(self):
        # Responsible for shutting down the machine
        raise NotImplementedError

    @defer.inlineCallbacks
    def substantiate(self, starting_worker):
        if self.state == States.STOPPING:
            # wait until stop action finishes
            yield self._stop_notifier.wait()

        if self.state == States.STARTED:
            # may happen if we waited for stop to complete and in the mean
            # time the machine was successfully woken.
            return True

        # wait for already proceeding startup to finish, if any
        if self.state == States.STARTING:
            return (yield self._start_notifier.wait())

        self.state = States.STARTING

        # substantiate all workers that will start if we wake the machine. We
        # do so before waking the machine to guarantee that we're already
        # waiting for worker connection as waking may take time confirming
        # machine came online. We'll call substantiate on the worker that
        # invoked this function again, but that's okay as that function is
        # reentrant. Note that we substantiate without gathering results
        # because the original call to substantiate will get them anyway and
        # we don't want to be slowed down by other workers on the machine.
        for worker in self.workers:
            if worker.starts_without_substantiate:
                worker.substantiate(None, None)

        # Start the machine. We don't need to wait for any workers to actually
        # come online as that's handled in their substantiate() functions.
        try:
            ret = yield self.start_machine()
        except Exception as e:
            log.err(e, f'while starting latent machine {self.name}')
            ret = False

        if not ret:
            yield defer.DeferredList(
                [worker.insubstantiate() for worker in self.workers], consumeErrors=True
            )
        else:
            self._setMissingTimer()

        self.state = States.STARTED if ret else States.STOPPED
        self._start_notifier.notify(ret)

        return ret

    @defer.inlineCallbacks
    def _stop(self):
        if any(worker.building for worker in self.workers) or self.state == States.STARTING:
            return None

        if self.state == States.STOPPING:
            yield self._stop_notifier.wait()
            return None

        self.state = States.STOPPING

        # wait until workers insubstantiate, then stop
        yield defer.DeferredList(
            [worker.insubstantiate() for worker in self.workers], consumeErrors=True
        )
        try:
            yield self.stop_machine()
        except Exception as e:
            log.err(e, f'while stopping latent machine {self.name}')

        self.state = States.STOPPED
        self._stop_notifier.notify(None)
        return None

    def notifyBuildStarted(self):
        self._clearMissingTimer()

    def notifyBuildFinished(self):
        if any(worker.building for worker in self.workers):
            self._clearBuildWaitTimer()
        else:
            self._setBuildWaitTimer()

    def _clearMissingTimer(self):
        if self._missing_timer is not None:
            if self._missing_timer.active():
                self._missing_timer.cancel()
            self._missing_timer = None

    def _setMissingTimer(self):
        self._clearMissingTimer()
        self._missing_timer = self.master.reactor.callLater(self.missing_timeout, self._stop)

    def _clearBuildWaitTimer(self):
        if self._build_wait_timer is not None:
            if self._build_wait_timer.active():
                self._build_wait_timer.cancel()
            self._build_wait_timer = None

    def _setBuildWaitTimer(self):
        self._clearBuildWaitTimer()
        self._build_wait_timer = self.master.reactor.callLater(self.build_wait_timeout, self._stop)

    def __repr__(self):
        return f"<AbstractLatentMachine '{self.name}' at {id(self)}>"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import collections
import json
import re
import weakref
from typing import Callable
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.python.components import registerAdapter
from zope.interface import implementer

from buildbot import config
from buildbot import util
from buildbot.interfaces import IProperties
from buildbot.interfaces import IRenderable
from buildbot.util import flatten


@implementer(IProperties)
class Properties(util.ComparableMixin):
    """
    I represent a set of properties that can be interpolated into various
    strings in buildsteps.

    @ivar properties: dictionary mapping property values to tuples
        (value, source), where source is a string identifying the source
        of the property.

    Objects of this class can be read like a dictionary -- in this case,
    only the property value is returned.

    As a special case, a property value of None is returned as an empty
    string when used as a mapping.
    """

    compare_attrs: ClassVar[Sequence[str]] = ('properties',)

    def __init__(self, **kwargs):
        """
        @param kwargs: initial property values (for testing)
        """
        self.properties = {}
        # Track keys which are 'runtime', and should not be
        # persisted if a build is rebuilt
        self.runtime = set()
        self.build = None  # will be set by the Build when starting
        self._used_secrets = {}
        if kwargs:
            self.update(kwargs, "TEST")
        self._master = None
        self._sourcestamps = None
        self._changes = None

    @property
    def master(self):
        if self.build is not None:
            return self.build.master
        return self._master

    @master.setter
    def master(self, value):
        self._master = value

    @property
    def sourcestamps(self):
        if self.build is not None:
            return [b.asDict() for b in self.build.getAllSourceStamps()]
        elif self._sourcestamps is not None:
            return self._sourcestamps
        raise AttributeError('neither build nor _sourcestamps are set')

    @sourcestamps.setter
    def sourcestamps(self, value):
        self._sourcestamps = value

    def getSourceStamp(self, codebase=''):
        for source in self.sourcestamps:
            if source['codebase'] == codebase:
                return source
        return None

    @property
    def changes(self):
        if self.build is not None:
            return [c.asChDict() for c in self.build.allChanges()]
        elif self._changes is not None:
            return self._changes
        raise AttributeError('neither build nor _changes are set')

    @changes.setter
    def changes(self, value):
        self._changes = value

    @property
    def files(self):
        if self.build is not None:
            return self.build.allFiles()
        files = []
        # self.changes, not self._changes to raise AttributeError if unset
        for chdict in self.changes:
            files.extend(chdict['files'])
        return files

    @classmethod
    def fromDict(cls, propDict):
        properties = cls()
        for name, (value, source) in propDict.items():
            properties.setProperty(name, value, source)
        return properties

    def __getstate__(self):
        d = self.__dict__.copy()
        d['build'] = None
        return d

    def __setstate__(self, d):
        self.__dict__ = d
        if not hasattr(self, 'runtime'):
            self.runtime = set()

    def __contains__(self, name):
        return name in self.properties

    def __getitem__(self, name):
        """Just get the value for this property."""
        rv = self.properties[name][0]
        return rv

    def __bool__(self):
        return bool(self.properties)

    def getPropertySource(self, name):
        return self.properties[name][1]

    def asList(self):
        """Return the properties as a sorted list of (name, value, source)"""
        ret = sorted([(k, v[0], v[1]) for k, v in self.properties.items()])
        return ret

    def asDict(self):
        """Return the properties as a simple key:value dictionary,
        properly unicoded"""
        return self.properties.copy()

    def __repr__(self):
        return 'Properties(**' + repr(dict((k, v[0]) for k, v in self.properties.items())) + ')'

    def update(self, dict, source, runtime=False):
        """Update this object from a dictionary, with an explicit source specified."""
        for k, v in dict.items():
            self.setProperty(k, v, source, runtime=runtime)

    def updateFromProperties(self, other):
        """Update this object based on another object; the other object's"""
        self.properties.update(other.properties)
        self.runtime.update(other.runtime)

    def updateFromPropertiesNoRuntime(self, other):
        """Update this object based on another object, but don't
        include properties that were marked as runtime."""
        for k, v in other.properties.items():
            if k not in other.runtime:
                self.properties[k] = v

    # IProperties methods

    def getProperty(self, name, default=None):
        return self.properties.get(name, (default,))[0]

    def hasProperty(self, name):
        return name in self.properties

    has_key = hasProperty

    def setProperty(self, name, value, source, runtime=False):
        name = util.bytes2unicode(name)
        if not IRenderable.providedBy(value):
            json.dumps(value)  # Let the exception propagate ...
        source = util.bytes2unicode(source)

        self.properties[name] = (value, source)
        if runtime:
            self.runtime.add(name)

    def getProperties(self):
        return self

    def getBuild(self):
        return self.build

    def render(self, value):
        renderable = IRenderable(value)
        return defer.maybeDeferred(renderable.getRenderingFor, self)

    # as the secrets are used in the renderable, they can pretty much arrive anywhere
    # in the log of state strings
    # so we have the renderable record here which secrets are used that we must remove
    def useSecret(self, secret_value, secret_name):
        if secret_value.strip():
            self._used_secrets[secret_value] = "<" + secret_name + ">"

    # This method shall then be called to remove secrets from any text that could be logged
    # somewhere and that could contain secrets
    def cleanupTextFromSecrets(self, text):
        # Better be correct and inefficient than efficient and wrong
        secrets = self._used_secrets
        for k in sorted(secrets, key=len, reverse=True):
            text = text.replace(k, secrets[k])
        return text


class PropertiesMixin:
    """
    A mixin to add L{IProperties} methods to a class which does not implement
    the full interface, only getProperties() function.

    This is useful because L{IProperties} methods are often called on L{Build}
    objects without first coercing them.

    @ivar set_runtime_properties: the default value for the C{runtime}
    parameter of L{setProperty}.
    """

    set_runtime_properties = False
    getProperties: Callable[[], IProperties]

    def getProperty(self, propname, default=None):
        return self.getProperties().getProperty(propname, default)

    def hasProperty(self, propname):
        return self.getProperties().hasProperty(propname)

    has_key = hasProperty

    def setProperty(self, propname, value, source='Unknown', runtime=None):
        # source is not optional in IProperties, but is optional here to avoid
        # breaking user-supplied code that fails to specify a source
        props = self.getProperties()
        if runtime is None:
            runtime = self.set_runtime_properties
        props.setProperty(propname, value, source, runtime=runtime)

    def render(self, value) -> defer.Deferred[IRenderable]:
        return self.getProperties().render(value)


@implementer(IRenderable)
class RenderableOperatorsMixin:
    """
    Properties and Interpolate instances can be manipulated with standard operators.
    """

    def __eq__(self, other):
        return _OperatorRenderer(self, other, "==", lambda v1, v2: v1 == v2)

    def __ne__(self, other):
        return _OperatorRenderer(self, other, "!=", lambda v1, v2: v1 != v2)

    def __lt__(self, other):
        return _OperatorRenderer(self, other, "<", lambda v1, v2: v1 < v2)

    def __le__(self, other):
        return _OperatorRenderer(self, other, "<=", lambda v1, v2: v1 <= v2)

    def __gt__(self, other):
        return _OperatorRenderer(self, other, ">", lambda v1, v2: v1 > v2)

    def __ge__(self, other):
        return _OperatorRenderer(self, other, ">=", lambda v1, v2: v1 >= v2)

    def __add__(self, other):
        return _OperatorRenderer(self, other, "+", lambda v1, v2: v1 + v2)

    def __sub__(self, other):
        return _OperatorRenderer(self, other, "-", lambda v1, v2: v1 - v2)

    def __mul__(self, other):
        return _OperatorRenderer(self, other, "*", lambda v1, v2: v1 * v2)

    def __truediv__(self, other):
        return _OperatorRenderer(self, other, "/", lambda v1, v2: v1 / v2)

    def __floordiv__(self, other):
        return _OperatorRenderer(self, other, "//", lambda v1, v2: v1 // v2)

    def __mod__(self, other):
        return _OperatorRenderer(self, other, "%", lambda v1, v2: v1 % v2)

    # we cannot use this trick to overload the 'in' operator, as python will force the result
    # of __contains__ to a boolean, forcing it to True all the time
    # so we mimic sqlalchemy and make a in_ method
    def in_(self, other):
        return _OperatorRenderer(self, other, "in", lambda v1, v2: v1 in v2)

    def getRenderingFor(self, iprops: IProperties) -> defer.Deferred:
        raise NotImplementedError


@implementer(IRenderable)
class _OperatorRenderer(RenderableOperatorsMixin, util.ComparableMixin):
    """
    An instance of this class renders a comparison given by a operator
    function with v1 and v2

    """

    compare_attrs: ClassVar[Sequence[str]] = ('fn',)

    def __init__(self, v1, v2, cstr, comparator):
        self.v1 = v1
        self.v2 = v2
        self.comparator = comparator
        self.cstr = cstr

    @defer.inlineCallbacks
    def getRenderingFor(self, props):
        v1, v2 = yield props.render((self.v1, self.v2))
        return self.comparator(v1, v2)

    def __repr__(self):
        return f'{self.v1!r} {self.cstr!s} {self.v2!r}'


class _PropertyMap:
    """
    Privately-used mapping object to implement WithProperties' substitutions,
    including the rendering of None as ''.
    """

    colon_minus_re = re.compile(r"(.*):-(.*)")
    colon_tilde_re = re.compile(r"(.*):~(.*)")
    colon_plus_re = re.compile(r"(.*):\+(.*)")

    def __init__(self, properties):
        # use weakref here to avoid a reference loop
        self.properties = weakref.ref(properties)
        self.temp_vals = {}

    def __getitem__(self, key):
        properties = self.properties()
        assert properties is not None

        def colon_minus(mo):
            # %(prop:-repl)s
            # if prop exists, use it; otherwise, use repl
            prop, repl = mo.group(1, 2)
            if prop in self.temp_vals:
                return self.temp_vals[prop]
            elif prop in properties:
                return properties[prop]
            return repl

        def colon_tilde(mo):
            # %(prop:~repl)s
            # if prop exists and is true (nonempty), use it; otherwise, use
            # repl
            prop, repl = mo.group(1, 2)
            if self.temp_vals.get(prop):
                return self.temp_vals[prop]
            elif prop in properties and properties[prop]:  # noqa: RUF019
                return properties[prop]
            return repl

        def colon_plus(mo):
            # %(prop:+repl)s
            # if prop exists, use repl; otherwise, an empty string
            prop, repl = mo.group(1, 2)
            if prop in properties or prop in self.temp_vals:
                return repl
            return ''

        for regexp, fn in [
            (self.colon_minus_re, colon_minus),
            (self.colon_tilde_re, colon_tilde),
            (self.colon_plus_re, colon_plus),
        ]:
            mo = regexp.match(key)
            if mo:
                rv = fn(mo)
                break
        else:
            # If explicitly passed as a kwarg, use that,
            # otherwise, use the property value.
            if key in self.temp_vals:
                rv = self.temp_vals[key]
            else:
                rv = properties[key]

        # translate 'None' to an empty string
        if rv is None:
            rv = ''
        return rv

    def add_temporary_value(self, key, val):
        "Add a temporary value (to support keyword arguments to WithProperties)"
        self.temp_vals[key] = val


@implementer(IRenderable)
class WithProperties(util.ComparableMixin):
    """
    This is a marker class, used fairly widely to indicate that we
    want to interpolate build properties.
    """

    compare_attrs: ClassVar[Sequence[str]] = ('fmtstring', 'args', 'lambda_subs')

    def __init__(self, fmtstring, *args, **lambda_subs):
        self.fmtstring = fmtstring
        self.args = args
        if not self.args:
            self.lambda_subs = lambda_subs
            for key, val in self.lambda_subs.items():
                if not callable(val):
                    raise ValueError(f'Value for lambda substitution "{key}" must be callable.')
        elif lambda_subs:
            raise ValueError(
                'WithProperties takes either positional or keyword substitutions, not both.'
            )

    def getRenderingFor(self, build):
        pmap = _PropertyMap(build.getProperties())
        if self.args:
            strings = []
            for name in self.args:
                strings.append(pmap[name])
            s = self.fmtstring % tuple(strings)
        else:
            for k, v in self.lambda_subs.items():
                pmap.add_temporary_value(k, v(build))
            s = self.fmtstring % pmap
        return s


class _NotHasKey(util.ComparableMixin):
    """A marker for missing ``hasKey`` parameter.

    To withstand ``deepcopy``, ``reload`` and pickle serialization round trips,
    check it with ``==`` or ``!=``.
    """

    compare_attrs: ClassVar[Sequence[str]] = ()


# any instance of _NotHasKey would do, yet we don't want to create and delete
# them all the time
_notHasKey = _NotHasKey()


@implementer(IRenderable)
class _Lookup(util.ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = (
        'value',
        'index',
        'default',
        'defaultWhenFalse',
        'hasKey',
        'elideNoneAs',
    )

    def __init__(
        self, value, index, default=None, defaultWhenFalse=True, hasKey=_notHasKey, elideNoneAs=None
    ):
        self.value = value
        self.index = index
        self.default = default
        self.defaultWhenFalse = defaultWhenFalse
        self.hasKey = hasKey
        self.elideNoneAs = elideNoneAs

    def __repr__(self):
        parts = [repr(self.index)]
        if self.default is not None:
            parts.append(f', default={self.default!r}')
        if not self.defaultWhenFalse:
            parts.append(', defaultWhenFalse=False')
        if self.hasKey != _notHasKey:
            parts.append(f', hasKey={self.hasKey!r}')
        if self.elideNoneAs is not None:
            parts.append(f', elideNoneAs={self.elideNoneAs!r}')

        parts_str = ''.join(parts)

        return f'_Lookup({self.value!r}, {parts_str})'

    @defer.inlineCallbacks
    def getRenderingFor(self, build):
        value = build.render(self.value)
        index = build.render(self.index)
        value, index = yield defer.gatherResults([value, index], consumeErrors=True)
        if index not in value:
            rv = yield build.render(self.default)
        else:
            if self.defaultWhenFalse:
                rv = yield build.render(value[index])
                if not rv:
                    rv = yield build.render(self.default)
                elif self.hasKey != _notHasKey:
                    rv = yield build.render(self.hasKey)
            elif self.hasKey != _notHasKey:
                rv = yield build.render(self.hasKey)
            else:
                rv = yield build.render(value[index])
        if rv is None:
            rv = yield build.render(self.elideNoneAs)
        return rv


def _getInterpolationList(fmtstring):
    # TODO: Verify that no positional substitutions are requested
    dd = collections.defaultdict(str)
    fmtstring % dd
    return list(dd)


@implementer(IRenderable)
class _PropertyDict:
    def getRenderingFor(self, build):
        return build.getProperties()


_thePropertyDict = _PropertyDict()


@implementer(IRenderable)
class _WorkerPropertyDict:
    def getRenderingFor(self, build):
        return build.getBuild().getWorkerInfo()


_theWorkerPropertyDict = _WorkerPropertyDict()


@implementer(IRenderable)
class _SecretRenderer:
    def __init__(self, secret_name):
        self.secret_name = secret_name

    @defer.inlineCallbacks
    def getRenderingFor(self, properties):
        secretsSrv = properties.master.namedServices.get("secrets")
        if not secretsSrv:
            error_message = (
                "secrets service not started, need to configure"
                " SecretManager in c['services'] to use 'secrets'"
                "in Interpolate"
            )
            raise KeyError(error_message)
        credsservice = properties.master.namedServices['secrets']
        secret_detail = yield credsservice.get(self.secret_name)
        if secret_detail is None:
            raise KeyError(f"secret key {self.secret_name} is not found in any provider")
        properties.useSecret(secret_detail.value, self.secret_name)
        return secret_detail.value


class Secret(_SecretRenderer):
    def __repr__(self):
        return f"Secret({self.secret_name})"


class _SecretIndexer:
    def __contains__(self, password):
        return True

    def __getitem__(self, password):
        return _SecretRenderer(password)


@implementer(IRenderable)
class _SourceStampDict(util.ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = ('codebase',)

    def __init__(self, codebase):
        self.codebase = codebase

    def getRenderingFor(self, props):
        ss = props.getSourceStamp(self.codebase)
        if ss:
            return ss
        return {}


@implementer(IRenderable)
class _Lazy(util.ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = ('value',)

    def __init__(self, value):
        self.value = value

    def getRenderingFor(self, build):
        return self.value

    def __repr__(self):
        return f'_Lazy({self.value!r})'


@implementer(IRenderable)
class Interpolate(RenderableOperatorsMixin, util.ComparableMixin):
    """
    This is a marker class, used fairly widely to indicate that we
    want to interpolate build properties.
    """

    compare_attrs: ClassVar[Sequence[str]] = ('fmtstring', 'args', 'kwargs')

    identifier_re = re.compile(r'^[\w._-]*$')

    def __init__(self, fmtstring, *args, **kwargs):
        self.fmtstring = fmtstring
        self.args = args
        self.kwargs = kwargs
        if self.args and self.kwargs:
            config.error("Interpolate takes either positional or keyword substitutions, not both.")
        if not self.args:
            self.interpolations = {}
            self._parse(fmtstring)

    def __repr__(self):
        if self.args:
            return f'Interpolate({self.fmtstring!r}, *{self.args!r})'
        elif self.kwargs:
            return f'Interpolate({self.fmtstring!r}, **{self.kwargs!r})'
        return f'Interpolate({self.fmtstring!r})'

    def _parse_substitution_prop(self, arg):
        try:
            prop, repl = arg.split(":", 1)
        except ValueError:
            prop = arg
            repl = None
        if not Interpolate.identifier_re.match(prop):
            config.error(f"Property name must be alphanumeric for prop Interpolation '{arg}'")
            prop = None
            repl = None

        return _thePropertyDict, prop, repl

    def _parse_substitution_secret(self, arg):
        try:
            secret, repl = arg.split(":", 1)
        except ValueError:
            secret = arg
            repl = None
        return _SecretIndexer(), secret, repl

    def _parse_substitution_src(self, arg):
        # TODO: Handle changes
        try:
            codebase, attr, repl = arg.split(":", 2)
        except ValueError:
            try:
                codebase, attr = arg.split(":", 1)
                repl = None
            except ValueError:
                config.error(
                    f"Must specify both codebase and attribute for src Interpolation '{arg}'"
                )
                return {}, None, None

        if not Interpolate.identifier_re.match(codebase):
            config.error(f"Codebase must be alphanumeric for src Interpolation '{arg}'")
            codebase = attr = repl = None
        if not Interpolate.identifier_re.match(attr):
            config.error(f"Attribute must be alphanumeric for src Interpolation '{arg}'")
            codebase = attr = repl = None
        return _SourceStampDict(codebase), attr, repl

    def _parse_substitution_worker(self, arg):
        try:
            prop, repl = arg.split(":", 1)
        except ValueError:
            prop = arg
            repl = None
        return _theWorkerPropertyDict, prop, repl

    def _parse_substitution_kw(self, arg):
        try:
            kw, repl = arg.split(":", 1)
        except ValueError:
            kw = arg
            repl = None
        if not Interpolate.identifier_re.match(kw):
            config.error(f"Keyword must be alphanumeric for kw Interpolation '{arg}'")
            kw = repl = None
        return _Lazy(self.kwargs), kw, repl

    _substitutions = {
        "prop": _parse_substitution_prop,
        "secret": _parse_substitution_secret,
        "src": _parse_substitution_src,
        "worker": _parse_substitution_worker,
        "kw": _parse_substitution_kw,
    }

    def _parseSubstitution(self, fmt):
        try:
            key, arg = fmt.split(":", 1)
        except ValueError:
            config.error(f"invalid Interpolate substitution without selector '{fmt}'")
            return None

        fn = self._substitutions.get(key, None)
        if not fn:
            config.error(f"invalid Interpolate selector '{key}'")
            return None
        return fn(self, arg)

    @staticmethod
    def _splitBalancedParen(delim, arg):
        parenCount = 0
        for i, val in enumerate(arg):
            if val == "(":
                parenCount += 1
            if val == ")":
                parenCount -= 1
                if parenCount < 0:
                    raise ValueError
            if parenCount == 0 and val == delim:
                return arg[0:i], arg[i + 1 :]
        return arg

    def _parseColon_minus(self, d, kw, repl):
        return _Lookup(
            d, kw, default=Interpolate(repl, **self.kwargs), defaultWhenFalse=False, elideNoneAs=''
        )

    def _parseColon_tilde(self, d, kw, repl):
        return _Lookup(
            d, kw, default=Interpolate(repl, **self.kwargs), defaultWhenFalse=True, elideNoneAs=''
        )

    def _parseColon_plus(self, d, kw, repl):
        return _Lookup(
            d,
            kw,
            hasKey=Interpolate(repl, **self.kwargs),
            default='',
            defaultWhenFalse=False,
            elideNoneAs='',
        )

    def _parseColon_ternary(self, d, kw, repl, defaultWhenFalse=False):
        delim = repl[0]
        if delim == '(':
            config.error("invalid Interpolate ternary delimiter '('")
            return None
        try:
            truePart, falsePart = self._splitBalancedParen(delim, repl[1:])
        except ValueError:
            config.error(
                f"invalid Interpolate ternary expression '{repl[1:]}' with delimiter '{repl[0]}'"
            )
            return None
        return _Lookup(
            d,
            kw,
            hasKey=Interpolate(truePart, **self.kwargs),
            default=Interpolate(falsePart, **self.kwargs),
            defaultWhenFalse=defaultWhenFalse,
            elideNoneAs='',
        )

    def _parseColon_ternary_hash(self, d, kw, repl):
        return self._parseColon_ternary(d, kw, repl, defaultWhenFalse=True)

    def _parse(self, fmtstring):
        keys = _getInterpolationList(fmtstring)
        for key in keys:
            if key not in self.interpolations:
                d, kw, repl = self._parseSubstitution(key)
                if repl is None:
                    repl = '-'
                for pattern, fn in [
                    ("-", self._parseColon_minus),
                    ("~", self._parseColon_tilde),
                    ("+", self._parseColon_plus),
                    ("?", self._parseColon_ternary),
                    ("#?", self._parseColon_ternary_hash),
                ]:
                    junk, matches, tail = repl.partition(pattern)
                    if not junk and matches:
                        self.interpolations[key] = fn(d, kw, tail)
                        break
                if key not in self.interpolations:
                    config.error(f"invalid Interpolate default type '{repl[0]}'")

    @defer.inlineCallbacks
    def getRenderingFor(self, build):
        props = build.getProperties()
        if self.args:
            args = yield props.render(self.args)
            return self.fmtstring % tuple(args)
        else:
            res = yield props.render(self.interpolations)
            return self.fmtstring % res


@implementer(IRenderable)
class Property(RenderableOperatorsMixin, util.ComparableMixin):
    """
    An instance of this class renders a property of a build.
    """

    compare_attrs: ClassVar[Sequence[str]] = ('key', 'default', 'defaultWhenFalse')

    def __init__(self, key, default=None, defaultWhenFalse=True):
        """
        @param key: Property to render.
        @param default: Value to use if property isn't set.
        @param defaultWhenFalse: When true (default), use default value
            if property evaluates to False. Otherwise, use default value
            only when property isn't set.
        """
        self.key = key
        self.default = default
        self.defaultWhenFalse = defaultWhenFalse

    def __repr__(self):
        return f"Property({self.key})"

    def getRenderingFor(self, props):
        if self.defaultWhenFalse:
            d = props.render(props.getProperty(self.key))

            @d.addCallback
            def checkDefault(rv):
                if rv:
                    return rv
                return props.render(self.default)

            return d

        if props.hasProperty(self.key):
            return props.render(props.getProperty(self.key))
        return props.render(self.default)


@implementer(IRenderable)
class FlattenList(RenderableOperatorsMixin, util.ComparableMixin):
    """
    An instance of this class flattens all nested lists in a list
    """

    compare_attrs: ClassVar[Sequence[str]] = ('nestedlist',)

    def __init__(self, nestedlist, types=(list, tuple)):
        """
        @param nestedlist: a list of values to render
        @param types: only flatten these types. defaults to (list, tuple)
        """
        self.nestedlist = nestedlist
        self.types = types

    @defer.inlineCallbacks
    def getRenderingFor(self, props):
        r = yield props.render(self.nestedlist)
        return flatten(r, self.types)

    def __add__(self, b):
        if isinstance(b, FlattenList):
            b = b.nestedlist
        return FlattenList(self.nestedlist + b, self.types)


@implementer(IRenderable)
class _Renderer(util.ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = ('fn',)

    def __init__(self, fn):
        self.fn = fn
        self.args = []
        self.kwargs = {}

    def withArgs(self, *args, **kwargs):
        new_renderer = _Renderer(self.fn)
        new_renderer.args = self.args + list(args)
        new_renderer.kwargs = dict(self.kwargs)
        new_renderer.kwargs.update(kwargs)
        return new_renderer

    @defer.inlineCallbacks
    def getRenderingFor(self, props):
        args = yield props.render(self.args)
        kwargs = yield props.render(self.kwargs)

        # We allow the renderer fn to return a renderable for convenience
        result = yield self.fn(props, *args, **kwargs)
        result = yield props.render(result)
        return result

    def __repr__(self):
        if self.args or self.kwargs:
            return f'renderer({self.fn!r}, args={self.args!r}, kwargs={self.kwargs!r})'
        return f'renderer({self.fn!r})'


def renderer(fn):
    return _Renderer(fn)


@implementer(IRenderable)
class _DefaultRenderer:
    """
    Default IRenderable adaptor. Calls .getRenderingFor if available, otherwise
    returns argument unchanged.
    """

    def __init__(self, value):
        try:
            self.renderer = value.getRenderingFor
        except AttributeError:
            self.renderer = lambda _: value

    def getRenderingFor(self, build):
        return self.renderer(build)


registerAdapter(_DefaultRenderer, object, IRenderable)


@implementer(IRenderable)
class _ListRenderer:
    """
    List IRenderable adaptor. Maps Build.render over the list.
    """

    def __init__(self, value):
        self.value = value

    def getRenderingFor(self, build):
        return defer.gatherResults([build.render(e) for e in self.value], consumeErrors=True)


registerAdapter(_ListRenderer, list, IRenderable)


@implementer(IRenderable)
class _TupleRenderer:
    """
    Tuple IRenderable adaptor. Maps Build.render over the tuple.
    """

    def __init__(self, value):
        self.value = value

    def getRenderingFor(self, build):
        d = defer.gatherResults([build.render(e) for e in self.value], consumeErrors=True)
        d.addCallback(tuple)
        return d


registerAdapter(_TupleRenderer, tuple, IRenderable)


@implementer(IRenderable)
class _DictRenderer:
    """
    Dict IRenderable adaptor. Maps Build.render over the keys and values in the dict.
    """

    def __init__(self, value):
        self.value = _ListRenderer([_TupleRenderer((k, v)) for k, v in value.items()])

    def getRenderingFor(self, build):
        d = self.value.getRenderingFor(build)
        d.addCallback(dict)
        return d


registerAdapter(_DictRenderer, dict, IRenderable)


@implementer(IRenderable)
class Transform:
    """
    A renderable that combines other renderables' results using an arbitrary function.
    """

    def __init__(self, function, *args, **kwargs):
        if not callable(function) and not IRenderable.providedBy(function):
            config.error("function given to Transform neither callable nor renderable")

        self._function = function
        self._args = args
        self._kwargs = kwargs

    @defer.inlineCallbacks
    def getRenderingFor(self, iprops):
        rfunction = yield iprops.render(self._function)
        rargs = yield iprops.render(self._args)
        rkwargs = yield iprops.render(self._kwargs)
        return rfunction(*rargs, **rkwargs)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import sys

from twisted.internet import protocol
from twisted.internet import reactor
from twisted.python.runtime import platformType

from buildbot.scripts import base
from buildbot.scripts.logwatcher import BuildmasterStartupError
from buildbot.scripts.logwatcher import BuildmasterTimeoutError
from buildbot.scripts.logwatcher import LogWatcher
from buildbot.scripts.logwatcher import ReconfigError
from buildbot.util import rewrap


class Follower:
    def follow(self, basedir, timeout=None):
        self.rc = 0
        self._timeout = timeout if timeout else 10.0
        print("Following twistd.log until startup finished..")
        lw = LogWatcher(os.path.join(basedir, "twistd.log"), timeout=self._timeout)
        d = lw.start()
        d.addCallbacks(self._success, self._failure)
        reactor.run()
        return self.rc

    def _success(self, _):
        print("The buildmaster appears to have (re)started correctly.")
        self.rc = 0
        reactor.stop()

    def _failure(self, why):
        if why.check(BuildmasterTimeoutError):
            print(
                rewrap(f"""\
                The buildmaster took more than {self._timeout} seconds to start, so we were
                unable to confirm that it started correctly.
                Please 'tail twistd.log' and look for a line that says
                'BuildMaster is running' to verify correct startup.
                """)
            )
        elif why.check(ReconfigError):
            print(
                rewrap("""\
                The buildmaster appears to have encountered an error in the
                master.cfg config file during startup.
                Please inspect and fix master.cfg, then restart the
                buildmaster.
                """)
            )
        elif why.check(BuildmasterStartupError):
            print(
                rewrap("""\
                The buildmaster startup failed. Please see 'twistd.log' for
                possible reason.
                """)
            )
        else:
            print(
                rewrap("""\
                Unable to confirm that the buildmaster started correctly.
                You may need to stop it, fix the config file, and restart.
                """)
            )
            print(why)
        self.rc = 1
        reactor.stop()


def launchNoDaemon(config):
    os.chdir(config['basedir'])
    sys.path.insert(0, os.path.abspath(config['basedir']))

    argv = [
        "twistd",
        "--no_save",
        "--nodaemon",
        "--logfile=twistd.log",  # windows doesn't use the same default
        "--python=buildbot.tac",
    ]

    if platformType != 'win32':
        # windows doesn't use pidfile option.
        argv.extend(["--pidfile="])

    sys.argv = argv

    # this is copied from bin/twistd. twisted-2.0.0 through 2.4.0 use
    # _twistw.run . Twisted-2.5.0 and later use twistd.run, even for
    # windows.
    from twisted.scripts import twistd

    twistd.run()


def launch(config):
    os.chdir(config['basedir'])
    sys.path.insert(0, os.path.abspath(config['basedir']))

    # see if we can launch the application without actually having to
    # spawn twistd, since spawning processes correctly is a real hassle
    # on windows.
    argv = [
        sys.executable,
        "-c",
        # this is copied from bin/twistd. twisted-2.0.0 through 2.4.0 use
        # _twistw.run . Twisted-2.5.0 and later use twistd.run, even for
        # windows.
        "from twisted.scripts import twistd; twistd.run()",
        "--no_save",
        "--logfile=twistd.log",  # windows doesn't use the same default
        "--python=buildbot.tac",
    ]

    # ProcessProtocol just ignores all output
    proc = reactor.spawnProcess(protocol.ProcessProtocol(), sys.executable, argv, env=os.environ)

    if platformType == "win32":
        with open("twistd.pid", "w", encoding='utf-8') as pidfile:
            pidfile.write(f"{proc.pid}")


def start(config):
    if not base.isBuildmasterDir(config['basedir']):
        return 1

    if config['nodaemon']:
        launchNoDaemon(config)
        return 0

    launch(config)

    # We don't have tail on windows
    if platformType == "win32" or config['quiet']:
        return 0

    # this is the parent
    timeout = config.get('start_timeout', None)
    if timeout is None:
        timeout = os.getenv('START_TIMEOUT', None)
    if timeout is not None:
        try:
            timeout = float(timeout)
        except ValueError:
            print('Start timeout must be a number')
            return 1

    rc = Follower().follow(config['basedir'], timeout=timeout)
    return rc



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import exceptions
from buildbot.data import types
from buildbot.util import identifiers

if TYPE_CHECKING:
    from buildbot.db.workers import WorkerModel
    from buildbot.util.twisted import InlineCallbacksType


def _db2data(model: WorkerModel):
    return {
        'workerid': model.id,
        'name': model.name,
        'workerinfo': model.workerinfo,
        'paused': model.paused,
        "pause_reason": model.pause_reason,
        'graceful': model.graceful,
        'connected_to': [{'masterid': id} for id in model.connected_to],
        'configured_on': [
            {'masterid': c.masterid, 'builderid': c.builderid} for c in model.configured_on
        ],
    }


class WorkerEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/workers/n:workerid",
        "/workers/i:name",
        "/masters/n:masterid/workers/n:workerid",
        "/masters/n:masterid/workers/i:name",
        "/masters/n:masterid/builders/n:builderid/workers/n:workerid",
        "/masters/n:masterid/builders/n:builderid/workers/i:name",
        "/builders/n:builderid/workers/n:workerid",
        "/builders/n:builderid/workers/i:name",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        sldict = yield self.master.db.workers.getWorker(
            workerid=kwargs.get('workerid'),
            name=kwargs.get('name'),
            masterid=kwargs.get('masterid'),
            builderid=kwargs.get('builderid'),
        )
        if sldict:
            return _db2data(sldict)
        return None

    @defer.inlineCallbacks
    def control(self, action, args, kwargs):
        if action not in ("stop", "pause", "unpause", "kill"):
            raise exceptions.InvalidControlException(f"action: {action} is not supported")

        worker = yield self.get(None, kwargs)
        if worker is not None:
            self.master.mq.produce(
                ("control", "worker", str(worker["workerid"]), action),
                {"reason": kwargs.get("reason", args.get("reason", "no reason"))},
            )
        else:
            raise exceptions.exceptions.InvalidPathError("worker not found")


class WorkersEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    rootLinkName = 'workers'
    pathPatterns = [
        "/workers",
        "/masters/n:masterid/workers",
        "/masters/n:masterid/builders/n:builderid/workers",
        "/builders/n:builderid/workers",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        paused = resultSpec.popBooleanFilter('paused')
        graceful = resultSpec.popBooleanFilter('graceful')
        workers_dicts = yield self.master.db.workers.getWorkers(
            builderid=kwargs.get('builderid'),
            masterid=kwargs.get('masterid'),
            paused=paused,
            graceful=graceful,
        )
        return [_db2data(w) for w in workers_dicts]


class MasterBuilderEntityType(types.Entity):
    masterid = types.Integer()
    builderid = types.Integer()


class MasterIdEntityType(types.Entity):
    masterid = types.Integer()


class Worker(base.ResourceType):
    name = "worker"
    plural = "workers"
    endpoints = [WorkerEndpoint, WorkersEndpoint]
    eventPathPatterns = [
        "/workers/:workerid",
    ]

    class EntityType(types.Entity):
        workerid = types.Integer()
        name = types.String()
        connected_to = types.List(of=MasterIdEntityType("master_id"))
        configured_on = types.List(of=MasterBuilderEntityType("master_builder"))
        workerinfo = types.JsonObject()
        paused = types.Boolean()
        pause_reason = types.NoneOk(types.String())
        graceful = types.Boolean()

    entityType = EntityType(name)

    @base.updateMethod
    # returns a Deferred that returns None
    def workerConfigured(self, workerid, masterid, builderids):
        return self.master.db.workers.workerConfigured(
            workerid=workerid, masterid=masterid, builderids=builderids
        )

    @base.updateMethod
    def findWorkerId(self, name: str) -> int:
        if not identifiers.isIdentifier(50, name):
            raise ValueError(f"Worker name {name!r} is not a 50-character identifier")
        return self.master.db.workers.findWorkerId(name)

    @base.updateMethod
    @defer.inlineCallbacks
    def workerConnected(
        self, workerid: int, masterid: int, workerinfo: str
    ) -> InlineCallbacksType[None]:
        yield self.master.db.workers.workerConnected(
            workerid=workerid, masterid=masterid, workerinfo=workerinfo
        )
        bs = yield self.master.data.get(('workers', workerid))
        self.produceEvent(bs, 'connected')

    @base.updateMethod
    @defer.inlineCallbacks
    def workerDisconnected(self, workerid: int, masterid: int) -> InlineCallbacksType[None]:
        yield self.master.db.workers.workerDisconnected(workerid=workerid, masterid=masterid)
        bs = yield self.master.data.get(('workers', workerid))
        self.produceEvent(bs, 'disconnected')

    @base.updateMethod
    @defer.inlineCallbacks
    def workerMissing(
        self, workerid: int, masterid: int, last_connection: int, notify: bool
    ) -> InlineCallbacksType[None]:
        bs = yield self.master.data.get(('workers', workerid))
        bs['last_connection'] = last_connection
        bs['notify'] = notify
        self.produceEvent(bs, 'missing')

    @base.updateMethod
    @defer.inlineCallbacks
    def set_worker_paused(
        self, workerid: int, paused: bool, pause_reason: str | None = None
    ) -> InlineCallbacksType[None]:
        yield self.master.db.workers.set_worker_paused(
            workerid=workerid, paused=paused, pause_reason=pause_reason
        )
        bs = yield self.master.data.get(('workers', workerid))
        self.produceEvent(bs, 'state_updated')

    @base.updateMethod
    @defer.inlineCallbacks
    def set_worker_graceful(self, workerid: int, graceful: bool) -> InlineCallbacksType[None]:
        yield self.master.db.workers.set_worker_graceful(workerid=workerid, graceful=graceful)
        bs = yield self.master.data.get(('workers', workerid))
        self.produceEvent(bs, 'state_updated')

    @base.updateMethod
    def deconfigureAllWorkersForMaster(self, masterid: int) -> defer.Deferred[None]:
        # unconfigure all workers for this master
        return self.master.db.workers.deconfigureAllWorkersForMaster(masterid=masterid)

    def _masterDeactivated(self, masterid: int) -> defer.Deferred[None]:
        return self.deconfigureAllWorkersForMaster(masterid)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Sequence
from typing import cast

from twisted.internet import defer
from twisted.logger import Logger

from buildbot import config
from buildbot.changes import base
from buildbot.util import bytes2unicode
from buildbot.util import datetime2epoch
from buildbot.util import httpclientservice
from buildbot.util.pullrequest import PullRequestMixin
from buildbot.util.state import StateMixin

if TYPE_CHECKING:
    from buildbot.util.twisted import InlineCallbacksType

log = Logger()

HOSTED_BASE_URL = "https://api.github.com"
link_urls = {"https": "clone_url", "svn": "svn_url", "git": "git_url", "ssh": "ssh_url"}


class GitHubPullrequestPoller(base.ReconfigurablePollingChangeSource, StateMixin, PullRequestMixin):
    compare_attrs: ClassVar[Sequence[str]] = (
        "owner",
        "repo",
        "token",
        "branches",
        "pollInterval",
        "category",
        "project",
        "pollAtLaunch",
        "name",
    )
    db_class_name = 'GitHubPullrequestPoller'
    property_basename = "github"

    def __init__(self, owner: str, repo: str, **kwargs: Any) -> None:
        name = kwargs.get("name")
        if not name:
            kwargs["name"] = "GitHubPullrequestPoller:" + owner + "/" + repo
        super().__init__(owner, repo, **kwargs)

    def checkConfig(  # type: ignore[override]
        self,
        owner: str,
        repo: str,
        branches: list[str] | None = None,
        category: str | Callable[[Any], Any] = 'pull',
        project: str | None = None,
        baseURL: str | None = None,
        pullrequest_filter: bool | Callable[[Any], bool] = True,
        token: str | None = None,
        magic_link: bool = False,
        repository_type: str = "https",
        github_property_whitelist: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        if repository_type not in ["https", "svn", "git", "ssh"]:
            config.error("repository_type must be one of {https, svn, git, ssh}")
        super().checkConfig(name=self.name, **kwargs)

    @defer.inlineCallbacks
    def reconfigService(  # type: ignore[override]
        self,
        owner: str,
        repo: str,
        branches: list[str] | None = None,
        pollInterval: int = 10 * 60,
        category: str | Callable[[Any], Any] | None = None,
        project: str | None = None,
        baseURL: str | None = None,
        pullrequest_filter: bool | Callable[[Any], bool] = True,
        token: str | None = None,
        pollAtLaunch: bool = False,
        magic_link: bool = False,
        repository_type: str = "https",
        github_property_whitelist: list[str] | None = None,
        **kwargs: Any,
    ) -> InlineCallbacksType[None]:
        yield super().reconfigService(name=self.name, **kwargs)

        if baseURL is None:
            baseURL = HOSTED_BASE_URL
        if baseURL.endswith('/'):
            baseURL = baseURL[:-1]

        http_headers = {'User-Agent': 'Buildbot'}
        if token is not None:
            token = yield self.renderSecrets(token)
            # token is not None here, but mypy doesn't know that
            http_headers.update({'Authorization': 'token ' + cast(str, token)})

        if github_property_whitelist is None:
            github_property_whitelist = []

        self._http = yield httpclientservice.HTTPSession(
            self.master.httpservice, baseURL, headers=http_headers
        )

        self.token = token
        self.owner = owner
        self.repo = repo
        self.branches = branches
        self.pollInterval = pollInterval
        self.pollAtLaunch = pollAtLaunch
        self.repository_type = link_urls[repository_type]
        self.magic_link = magic_link
        self.external_property_whitelist = github_property_whitelist

        if callable(pullrequest_filter):
            self.pullrequest_filter = pullrequest_filter
        else:
            self.pullrequest_filter = lambda _: pullrequest_filter

        self.category = category if callable(category) else bytes2unicode(category)
        self.project = bytes2unicode(project)

    def describe(self) -> str:
        return f"GitHubPullrequestPoller watching the GitHub repository {self.owner}/{self.repo}"

    @defer.inlineCallbacks
    def _getPullInformation(self, pull_number: int) -> InlineCallbacksType[dict[str, Any]]:
        result = yield self._http.get(
            '/'.join(['/repos', self.owner, self.repo, 'pulls', str(pull_number)])
        )
        my_json = yield result.json()
        return my_json

    @defer.inlineCallbacks
    def _getPulls(self) -> InlineCallbacksType[list[dict[str, Any]]]:
        log.debug(
            "GitHubPullrequestPoller: polling "
            f"GitHub repository {self.owner}/{self.repo}, branches: {self.branches}"
        )
        result = yield self._http.get('/'.join(['/repos', self.owner, self.repo, 'pulls']))
        my_json = yield result.json()
        if result.code != 200:
            message = my_json.get('message', 'unknown')
            log.error(
                f"GitHubPullrequestPoller error {result.code} '{message}' "
                f"while loading {result.url}"
            )
            return []
        return my_json

    @defer.inlineCallbacks
    def _getFiles(self, prnumber: int) -> InlineCallbacksType[list[str]]:
        result = yield self._http.get(
            "/".join(['/repos', self.owner, self.repo, 'pulls', str(prnumber), 'files'])
        )
        my_json = yield result.json()

        return [f["filename"] for f in my_json]

    @defer.inlineCallbacks
    def _getCommitters(self, prnumber: int) -> InlineCallbacksType[list[list[str]]]:
        result = yield self._http.get(
            "/".join(['/repos', self.owner, self.repo, 'pulls', str(prnumber), 'commits'])
        )
        my_json = yield result.json()

        return [
            [c["commit"]["committer"]["name"], c["commit"]["committer"]["email"]] for c in my_json
        ]

    @defer.inlineCallbacks
    def _getAuthors(self, prnumber: int) -> InlineCallbacksType[list[list[str]]]:
        result = yield self._http.get(
            "/".join(['/repos', self.owner, self.repo, 'pulls', str(prnumber), 'commits'])
        )
        my_json = yield result.json()

        return [[a["commit"]["author"]["name"], a["commit"]["author"]["email"]] for a in my_json]

    @defer.inlineCallbacks
    def _getCurrentRev(self, prnumber: int) -> InlineCallbacksType[str | None]:
        # Get currently assigned revision of PR number

        result = yield self._getStateObjectId()
        rev = yield self.master.db.state.getState(result, f'pull_request{prnumber}', None)
        return rev

    @defer.inlineCallbacks
    def _setCurrentRev(self, prnumber: int, rev: str) -> InlineCallbacksType[None]:
        # Set the updated revision for PR number.

        result = yield self._getStateObjectId()
        yield self.master.db.state.setState(result, f'pull_request{prnumber}', rev)

    @defer.inlineCallbacks
    def _getStateObjectId(self) -> InlineCallbacksType[int]:
        # Return a deferred for object id in state db.
        result = yield self.master.db.state.getObjectId(
            f'{self.owner}/{self.repo}', self.db_class_name
        )
        return result

    @defer.inlineCallbacks
    def _processChanges(self, github_result: list[dict[str, Any]]) -> InlineCallbacksType[None]:
        for pr in github_result:
            # Track PRs for specified branches
            base_branch = pr['base']['ref']
            prnumber = pr['number']
            revision = pr['head']['sha']

            # Check to see if the branch is set or matches
            if self.branches is not None and base_branch not in self.branches:
                continue
            if self.pullrequest_filter is not None and not self.pullrequest_filter(pr):
                continue
            current = yield self._getCurrentRev(prnumber)
            if not current or current[0:12] != revision[0:12]:
                # Access title, repo, html link, and comments
                pr = yield self._getPullInformation(prnumber)
                title = pr['title']
                if self.magic_link:
                    branch = f'refs/pull/{prnumber}/merge'
                    repo = pr['base']['repo'][self.repository_type]
                else:
                    branch = pr['head']['ref']
                    repo = pr['head']['repo'][self.repository_type]
                revlink = pr['html_url']
                comments = pr['body']
                updated = datetime.strptime(pr['updated_at'], '%Y-%m-%dT%H:%M:%SZ')
                # update database
                yield self._setCurrentRev(prnumber, revision)

                project = self.project
                if project is None:
                    project = pr['base']['repo']['full_name']
                commits = pr['commits']

                dl: defer.Deferred[Any] = defer.DeferredList(
                    [
                        self._getAuthors(prnumber),
                        self._getCommitters(prnumber),
                        self._getFiles(prnumber),
                    ],
                    consumeErrors=True,
                )

                results = yield dl
                failures = [r[1] for r in results if not r[0]]
                if failures:
                    for failure in failures:
                        log.error(
                            "while processing changes for "
                            f"Pullrequest {prnumber} revision {revision}: {failure}"
                        )
                    # Fail on the first error!
                    failures[0].raiseException()
                [authors, committers, files] = [r[1] for r in results]

                author = authors[0][0] + " <" + authors[0][1] + ">"

                committer = committers[0][0] + " <" + committers[0][1] + ">"

                # emit the change
                yield self.master.data.updates.addChange(
                    author=author,
                    committer=committer,
                    revision=bytes2unicode(revision),
                    revlink=bytes2unicode(revlink),
                    comments=f"GitHub Pull Request #{prnumber} "
                    f"({commits} commit{'s' if commits > 0 else ''})\n{title}\n{comments}",
                    when_timestamp=datetime2epoch(updated),
                    branch=bytes2unicode(branch),
                    category=self.category,
                    project=project,
                    repository=bytes2unicode(repo),
                    files=files,
                    properties={
                        'pullrequesturl': revlink,
                        **self.extractProperties(pr),
                    },
                    src='git',
                )

    @defer.inlineCallbacks
    def poll(self) -> InlineCallbacksType[None]:  # type: ignore[override]
        result = yield self._getPulls()
        yield self._processChanges(result)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
BuildSteps that are specific to the Twisted source tree
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.python import log

from buildbot import util
from buildbot.process import buildstep
from buildbot.process import logobserver
from buildbot.process.results import FAILURE
from buildbot.process.results import SKIPPED
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.steps import shell

if TYPE_CHECKING:
    from twisted.internet.base import ReactorBase


class HLint(buildstep.ShellMixin, buildstep.BuildStep):
    """I run a 'lint' checker over a set of .xhtml files. Any deviations
    from recommended style is flagged and put in the output log.

    This step looks at .changes in the parent Build to extract a list of
    Lore XHTML files to check."""

    name = "hlint"
    description = "running hlint"
    descriptionDone = "hlint"
    warnOnWarnings = True
    warnOnFailure = True
    # TODO: track time, but not output
    warnings = 0

    def __init__(self, python=None, **kwargs):
        kwargs = self.setupShellMixin(kwargs, prohibitArgs=['command'])
        super().__init__(**kwargs)
        self.python = python
        self.warningLines = []
        self.addLogObserver('stdio', logobserver.LineConsumerLogObserver(self.logConsumer))

    @defer.inlineCallbacks
    def run(self):
        # create the command
        html_files = set()
        for f in self.build.allFiles():
            if f.endswith(".xhtml") and not f.startswith("sandbox/"):
                html_files.add(f)
        # remove duplicates
        hlintTargets = sorted(list(html_files))
        if not hlintTargets:
            return SKIPPED
        self.hlintFiles = hlintTargets

        command = []
        if self.python:
            command.append(self.python)
        command += ["bin/lore", "-p", "--output", "lint", *self.hlintFiles]

        cmd = yield self.makeRemoteShellCommand(command=command)
        yield self.runCommand(cmd)

        stdio_log = yield self.getLog('stdio')
        yield stdio_log.finish()

        yield self.addCompleteLog('warnings', '\n'.join(self.warningLines))
        yield self.addCompleteLog("files", "\n".join(self.hlintFiles) + "\n")

        # warnings are in stdout, rc is always 0, unless the tools break
        if cmd.didFail():
            return FAILURE

        self.descriptionDone = f"{self.warnings} hlin{(self.warnings == 1 and 't') or 'ts'}"

        if self.warnings:
            return WARNINGS
        return SUCCESS

    def logConsumer(self):
        while True:
            _, line = yield
            if ':' in line:
                self.warnings += 1
                self.warningLines.append(line)


class TrialTestCaseCounter(logobserver.LogLineObserver):
    _line_re = re.compile(r'^(?:Doctest: )?([\w\.]+) \.\.\. \[([^\]]+)\]$')

    def __init__(self):
        super().__init__()
        self.numTests = 0
        self.finished = False
        self.counts = {
            'total': None,
            'failures': 0,
            'errors': 0,
            'skips': 0,
            'expectedFailures': 0,
            'unexpectedSuccesses': 0,
        }

    def outLineReceived(self, line):
        # different versions of Twisted emit different per-test lines with
        # the bwverbose reporter.
        #  2.0.0: testSlave (buildbot.test.test_runner.Create) ... [OK]
        #  2.1.0: buildbot.test.test_runner.Create.testSlave ... [OK]
        #  2.4.0: buildbot.test.test_runner.Create.testSlave ... [OK]
        # Let's just handle the most recent version, since it's the easiest.
        # Note that doctests create lines line this:
        #  Doctest: viff.field.GF ... [OK]

        if line.startswith("=" * 40):
            self.finished = True
        if not self.finished:
            m = self._line_re.search(line.strip())
            if m:
                m.groups()
                self.numTests += 1
                self.step.setProgress('tests', self.numTests)

        out = re.search(r'Ran (\d+) tests', line)
        if out:
            self.counts['total'] = int(out.group(1))
        if line.startswith("OK") or line.startswith("FAILED ") or line.startswith("PASSED"):
            # the extra space on FAILED_ is to distinguish the overall
            # status from an individual test which failed. The lack of a
            # space on the OK is because it may be printed without any
            # additional text (if there are no skips,etc)
            out = re.search(r'failures=(\d+)', line)
            if out:
                self.counts['failures'] = int(out.group(1))
            out = re.search(r'errors=(\d+)', line)
            if out:
                self.counts['errors'] = int(out.group(1))
            out = re.search(r'skips=(\d+)', line)
            if out:
                self.counts['skips'] = int(out.group(1))
            out = re.search(r'expectedFailures=(\d+)', line)
            if out:
                self.counts['expectedFailures'] = int(out.group(1))
            out = re.search(r'unexpectedSuccesses=(\d+)', line)
            if out:
                self.counts['unexpectedSuccesses'] = int(out.group(1))
            # successes= is a Twisted-2.0 addition, and is not currently used
            out = re.search(r'successes=(\d+)', line)
            if out:
                self.counts['successes'] = int(out.group(1))


UNSPECIFIED = ()  # since None is a valid choice


class Trial(buildstep.ShellMixin, buildstep.BuildStep):
    """
    There are some class attributes which may be usefully overridden
    by subclasses. 'trialMode' and 'trialArgs' can influence the trial
    command line.
    """

    name = "trial"
    progressMetrics = ('output', 'tests', 'test.log')
    # note: the slash only works on unix workers, of course, but we have
    # no way to know what the worker uses as a separator.
    # TODO: figure out something clever.
    logfiles = {"test.log": "_trial_temp/test.log"}
    # we use test.log to track Progress at the end of __init__()

    renderables = ['tests', 'jobs']
    flunkOnFailure = True
    python: list[str] | str | None = None
    trial = "trial"
    trialMode = ["--reporter=bwverbose"]  # requires Twisted-2.1.0 or newer
    # for Twisted-2.0.0 or 1.3.0, use ["-o"] instead
    trialArgs: list[str] = []
    jobs: int | None = None
    testpath = UNSPECIFIED  # required (but can be None)
    testChanges = False  # TODO: needs better name
    recurse = False
    reactor: ReactorBase | None = None
    randomly = False
    tests: list[str] | None = None  # required

    description = 'testing'
    descriptionDone = 'tests'

    def __init__(
        self,
        reactor=UNSPECIFIED,
        python=None,
        trial=None,
        testpath=UNSPECIFIED,
        tests=None,
        testChanges=None,
        recurse=None,
        randomly=None,
        trialMode=None,
        trialArgs=None,
        jobs=None,
        **kwargs,
    ):
        kwargs = self.setupShellMixin(kwargs, prohibitArgs=['command'])
        super().__init__(**kwargs)

        if python:
            self.python = python
        if self.python is not None:
            if isinstance(self.python, str):
                self.python = [self.python]
            for s in self.python:
                if " " in s:
                    # this is not strictly an error, but I suspect more
                    # people will accidentally try to use python="python2.3
                    # -Wall" than will use embedded spaces in a python flag
                    log.msg("python= component '%s' has spaces")
                    log.msg("To add -Wall, use python=['python', '-Wall']")
                    why = "python= value has spaces, probably an error"
                    raise ValueError(why)

        if trial:
            self.trial = trial
        if " " in self.trial:
            raise ValueError("trial= value has spaces")
        if trialMode is not None:
            self.trialMode = trialMode
        if trialArgs is not None:
            self.trialArgs = trialArgs
        if jobs is not None:
            self.jobs = jobs

        if testpath is not UNSPECIFIED:
            self.testpath = testpath
        if self.testpath is UNSPECIFIED:
            raise ValueError("You must specify testpath= (it can be None)")
        assert isinstance(self.testpath, str) or self.testpath is None

        if reactor is not UNSPECIFIED:
            self.reactor = reactor

        if tests is not None:
            self.tests = tests
        if isinstance(self.tests, str):
            self.tests = [self.tests]
        if testChanges is not None:
            self.testChanges = testChanges
            # self.recurse = True  # not sure this is necessary

        if not self.testChanges and self.tests is None:
            raise ValueError("Must either set testChanges= or provide tests=")

        if recurse is not None:
            self.recurse = recurse
        if randomly is not None:
            self.randomly = randomly

        if self.reactor:
            self.description = f"testing ({self.reactor})"

        # this counter will feed Progress along the 'test cases' metric
        self.observer = TrialTestCaseCounter()
        self.addLogObserver('stdio', self.observer)

        # this observer consumes multiple lines in a go, so it can't be easily
        # handled in TrialTestCaseCounter.
        self.addLogObserver('stdio', logobserver.LineConsumerLogObserver(self.logConsumer))
        self.problems = []
        self.warnings = {}

        # text used before commandComplete runs
        self.text = 'running'

    def setup_python_path(self):
        if self.testpath is None:
            return

        # this bit produces a list, which can be used by buildbot_worker.runprocess.RunProcess
        ppath = self.env.get('PYTHONPATH', self.testpath)
        if isinstance(ppath, str):
            ppath = [ppath]
        if self.testpath not in ppath:
            ppath.insert(0, self.testpath)
        self.env['PYTHONPATH'] = ppath

    @defer.inlineCallbacks
    def run(self):
        # choose progressMetrics and logfiles based on whether trial is being
        # run with multiple workers or not.
        output_observer = logobserver.OutputProgressObserver('test.log')

        # build up most of the command, then stash it until start()
        command = []
        if self.python:
            command.extend(self.python)
        command.append(self.trial)
        command.extend(self.trialMode)
        if self.recurse:
            command.append("--recurse")
        if self.reactor:
            command.append(f"--reactor={self.reactor}")
        if self.randomly:
            command.append("--random=0")
        command.extend(self.trialArgs)

        if self.jobs is not None:
            self.jobs = int(self.jobs)
            command.append(f"--jobs={self.jobs}")

            # using -j/--jobs flag produces more than one test log.
            self.logfiles = {}
            for i in range(self.jobs):
                self.logfiles[f'test.{i}.log'] = f'_trial_temp/{i}/test.log'
                self.logfiles[f'err.{i}.log'] = f'_trial_temp/{i}/err.log'
                self.logfiles[f'out.{i}.log'] = f'_trial_temp/{i}/out.log'
                self.addLogObserver(f'test.{i}.log', output_observer)
        else:
            # this one just measures bytes of output in _trial_temp/test.log
            self.addLogObserver('test.log', output_observer)

        # now that self.build.allFiles() is nailed down, finish building the
        # command
        if self.testChanges:
            for f in self.build.allFiles():
                if f.endswith(".py"):
                    command.append(f"--testmodule={f}")
        else:
            command.extend(self.tests)

        self.setup_python_path()

        cmd = yield self.makeRemoteShellCommand(command=command)
        yield self.runCommand(cmd)

        stdio_log = yield self.getLog('stdio')
        yield stdio_log.finish()

        # figure out all status, then let the various hook functions return
        # different pieces of it

        problems = '\n'.join(self.problems)
        warnings = self.warnings

        if problems:
            yield self.addCompleteLog("problems", problems)

        if warnings:
            lines = sorted(warnings.keys())
            yield self.addCompleteLog("warnings", "".join(lines))

        return self.build_results(cmd)

    def build_results(self, cmd):
        counts = self.observer.counts
        total = counts['total']
        failures = counts['failures']
        errors = counts['errors']
        parsed = total is not None

        desc_parts = []

        if not cmd.didFail():
            if parsed:
                results = SUCCESS
                if total:
                    desc_parts += [str(total), (total == 1 and "test") or "tests", "passed"]
                else:
                    desc_parts += ["no tests", "run"]
            else:
                results = FAILURE
                desc_parts += ["testlog", "unparseable"]
        else:
            # something failed
            results = FAILURE
            if parsed:
                desc_parts += ["tests"]
                if failures:
                    desc_parts += [str(failures), (failures == 1 and "failure") or "failures"]
                if errors:
                    desc_parts += [str(errors), (errors == 1 and "error") or "errors"]
            else:
                desc_parts += ["tests", "failed"]

        if counts['skips']:
            desc_parts += [str(counts['skips']), (counts['skips'] == 1 and "skip") or "skips"]
        if counts['expectedFailures']:
            desc_parts += [
                str(counts['expectedFailures']),
                "todo" if counts['expectedFailures'] == 1 else "todos",
            ]

        if self.reactor:
            desc_parts.append(self.rtext('({})'))

        self.descriptionDone = util.join_list(desc_parts)
        return results

    def rtext(self, fmt='{}'):
        if self.reactor:
            rtext = fmt.format(self.reactor)
            return rtext.replace("reactor", "")
        return ""

    def logConsumer(self):
        while True:
            _, line = yield
            if line.find(" exceptions.DeprecationWarning: ") != -1:
                # no source
                warning = line  # TODO: consider stripping basedir prefix here
                self.warnings[warning] = self.warnings.get(warning, 0) + 1
            elif line.find(" DeprecationWarning: ") != -1 or line.find(" UserWarning: ") != -1:
                # next line is the source
                warning = line + "\n" + (yield)[1] + "\n"
                self.warnings[warning] = self.warnings.get(warning, 0) + 1
            elif line.find("Warning: ") != -1:
                warning = line
                self.warnings[warning] = self.warnings.get(warning, 0) + 1

            if line.find("=" * 60) == 0 or line.find("-" * 60) == 0:
                # read to EOF
                while True:
                    self.problems.append(line)
                    _, line = yield


class RemovePYCs(shell.ShellCommand):
    name = "remove_pyc"
    command = ['find', '.', '-name', "'*.pyc'", '-exec', 'rm', '{}', ';']
    description = "removing .pyc files"
    descriptionDone = "remove .pycs"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class BuildRequest(Row):
    table = "buildrequests"

    id_column = 'id'

    def __init__(
        self,
        id=None,
        buildsetid=None,
        builderid=None,
        priority=0,
        complete=0,
        results=-1,
        submitted_at=12345678,
        complete_at=None,
        waited_for=0,
    ):
        super().__init__(
            id=id,
            buildsetid=buildsetid,
            builderid=builderid,
            priority=priority,
            complete=complete,
            results=results,
            submitted_at=submitted_at,
            complete_at=complete_at,
            waited_for=waited_for,
        )


class BuildRequestClaim(Row):
    table = "buildrequest_claims"

    def __init__(self, brid=None, masterid=None, claimed_at=None):
        super().__init__(brid=brid, masterid=masterid, claimed_at=claimed_at)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.db import base
from buildbot.util import epoch2datetime
from buildbot.util.twisted import async_to_deferred
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    import datetime


@dataclass
class UrlModel:
    name: str
    url: str

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'StepsConnectorComponent '
                'getStep, and getSteps '
                'no longer return Step as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@dataclass
class StepModel:
    id: int
    number: int
    name: str
    buildid: int
    started_at: datetime.datetime | None
    locks_acquired_at: datetime.datetime | None
    complete_at: datetime.datetime | None
    state_string: str
    results: int | None
    urls: list[UrlModel]
    hidden: bool = False

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'StepsConnectorComponent '
                'getStep, and getSteps '
                'no longer return Step as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class StepsConnectorComponent(base.DBConnectorComponent):
    url_lock: defer.DeferredLock | None = None

    @async_to_deferred
    async def getStep(
        self,
        stepid: int | None = None,
        buildid: int | None = None,
        number: int | None = None,
        name: str | None = None,
    ) -> StepModel | None:
        tbl = self.db.model.steps
        if stepid is not None:
            wc = tbl.c.id == stepid
        else:
            if buildid is None:
                raise RuntimeError('must supply either stepid or buildid')
            if number is not None:
                wc = tbl.c.number == number
            elif name is not None:
                wc = tbl.c.name == name
            else:
                raise RuntimeError('must supply either number or name')
            wc = wc & (tbl.c.buildid == buildid)

        def thd(conn) -> StepModel | None:
            q = self.db.model.steps.select().where(wc)
            res = conn.execute(q)
            row = res.fetchone()

            rv = None
            if row:
                rv = self._model_from_row(row)
            res.close()
            return rv

        return await self.db.pool.do(thd)

    def getSteps(self, buildid: int) -> defer.Deferred[list[StepModel]]:
        def thd(conn) -> list[StepModel]:
            tbl = self.db.model.steps
            q = tbl.select()
            q = q.where(tbl.c.buildid == buildid)
            q = q.order_by(tbl.c.number)
            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    def addStep(
        self, buildid: int, name: str, state_string: str
    ) -> defer.Deferred[tuple[int, int, str]]:
        def thd(conn) -> tuple[int, int, str]:
            tbl = self.db.model.steps
            # get the highest current number
            r = conn.execute(sa.select(sa.func.max(tbl.c.number)).where(tbl.c.buildid == buildid))
            number = r.scalar()
            number = 0 if number is None else number + 1

            # note that there is no chance for a race condition here,
            # since only one master is inserting steps.  If there is a
            # conflict, then the name is likely already taken.
            insert_row = {
                "buildid": buildid,
                "number": number,
                "started_at": None,
                "locks_acquired_at": None,
                "complete_at": None,
                "state_string": state_string,
                "urls_json": '[]',
                "name": name,
            }
            try:
                r = conn.execute(self.db.model.steps.insert(), insert_row)
                conn.commit()
                got_id = r.inserted_primary_key[0]
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                conn.rollback()
                got_id = None

            if got_id:
                return (got_id, number, name)

            # we didn't get an id, so calculate a unique name and use that
            # instead.  Because names are truncated at the right to fit in a
            # 50-character identifier, this isn't a simple query.
            res = conn.execute(sa.select(tbl.c.name).where(tbl.c.buildid == buildid))
            names = {row[0] for row in res}
            num = 1
            while True:
                numstr = f'_{num}'
                newname = name[: 50 - len(numstr)] + numstr
                if newname not in names:
                    break
                num += 1
            insert_row['name'] = newname
            r = conn.execute(self.db.model.steps.insert(), insert_row)
            conn.commit()
            got_id = r.inserted_primary_key[0]
            return (got_id, number, newname)

        return self.db.pool.do(thd)

    def startStep(self, stepid: int, started_at: int, locks_acquired: bool) -> defer.Deferred[None]:
        def thd(conn) -> None:
            tbl = self.db.model.steps
            q = tbl.update().where(tbl.c.id == stepid)
            if locks_acquired:
                conn.execute(q.values(started_at=started_at, locks_acquired_at=started_at))
            else:
                conn.execute(q.values(started_at=started_at))

        return self.db.pool.do_with_transaction(thd)

    def set_step_locks_acquired_at(
        self, stepid: int, locks_acquired_at: int
    ) -> defer.Deferred[None]:
        def thd(conn) -> None:
            tbl = self.db.model.steps
            q = tbl.update().where(tbl.c.id == stepid)
            conn.execute(q.values(locks_acquired_at=locks_acquired_at))

        return self.db.pool.do_with_transaction(thd)

    def setStepStateString(self, stepid: int, state_string: str) -> defer.Deferred[None]:
        def thd(conn) -> None:
            tbl = self.db.model.steps
            q = tbl.update().where(tbl.c.id == stepid)
            conn.execute(q.values(state_string=state_string))

        return self.db.pool.do_with_transaction(thd)

    def addURL(self, stepid: int, name: str, url: str, _racehook=None) -> defer.Deferred[None]:
        # This methods adds an URL to the db
        # This is a read modify write and thus there is a possibility
        # that several urls are added at the same time (e.g with a deferredlist
        # at the end of a step)
        # this race condition is only inside the same master, as only one master
        # is supposed to add urls to a buildstep.
        # so threading.lock is used, as we are in the thread pool
        if self.url_lock is None:
            # this runs in reactor thread, so no race here..
            self.url_lock = defer.DeferredLock()

        def thd(conn) -> None:
            tbl = self.db.model.steps
            wc = tbl.c.id == stepid
            q = sa.select(tbl.c.urls_json).where(wc)
            res = conn.execute(q)
            row = res.fetchone()
            if _racehook is not None:
                _racehook()
            urls = json.loads(row.urls_json)

            url_item = {"name": name, "url": url}

            if url_item not in urls:
                urls.append(url_item)
                q2 = tbl.update().where(wc)
                conn.execute(q2.values(urls_json=json.dumps(urls)))
                conn.commit()

        return self.url_lock.run(self.db.pool.do, thd)

    def finishStep(self, stepid: int, results: int, hidden: bool) -> defer.Deferred[None]:
        def thd(conn) -> None:
            tbl = self.db.model.steps
            q = tbl.update().where(tbl.c.id == stepid)
            conn.execute(
                q.values(
                    complete_at=int(self.master.reactor.seconds()),
                    results=results,
                    hidden=1 if hidden else 0,
                )
            )

        return self.db.pool.do_with_transaction(thd)

    def _model_from_row(self, row):
        return StepModel(
            id=row.id,
            number=row.number,
            name=row.name,
            buildid=row.buildid,
            started_at=epoch2datetime(row.started_at),
            locks_acquired_at=epoch2datetime(row.locks_acquired_at),
            complete_at=epoch2datetime(row.complete_at),
            state_string=row.state_string,
            results=row.results,
            urls=[UrlModel(item['name'], item['url']) for item in json.loads(row.urls_json)],
            hidden=bool(row.hidden),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import json
import uuid
from typing import TYPE_CHECKING
from typing import Any

from twisted.python import log
from twisted.web import resource
from twisted.web import server

from buildbot.data.exceptions import InvalidPathError
from buildbot.util import bytes2unicode
from buildbot.util import toJson
from buildbot.util import unicode2bytes

if TYPE_CHECKING:
    from buildbot.master import BuildMaster
    from buildbot.mq.base import QueueRef


class Consumer:
    qrefs: dict[bytes, QueueRef]

    def __init__(self, request: server.Request):
        self.request = request
        self.qrefs = {}

    def stopConsuming(self, key: bytes | None = None) -> None:
        if key is not None:
            self.qrefs[key].stopConsuming()
        else:
            for qref in self.qrefs.values():
                qref.stopConsuming()
            self.qrefs = {}

    def onMessage(self, event: list[str], data: Any) -> None:
        request = self.request
        key = [bytes2unicode(e) for e in event]
        msg = {"key": key, "message": data}
        request.write(b"event: " + b"event" + b"\n")
        request.write(b"data: " + unicode2bytes(json.dumps(msg, default=toJson)) + b"\n")
        request.write(b"\n")

    def registerQref(self, path: bytes, qref: QueueRef) -> None:
        self.qrefs[path] = qref


class EventResource(resource.Resource):
    isLeaf = True
    consumers: dict[bytes, Consumer]

    def __init__(self, master: BuildMaster):
        super().__init__()

        self.master = master
        self.consumers = {}

    def decodePath(self, path: list[bytes]) -> list[bytes | None]:
        return [None if p == b'*' else p for p in path]

    def finish(self, request: server.Request, code: int, msg: bytes) -> None:
        request.setResponseCode(code)
        request.setHeader(b'content-type', b'text/plain; charset=utf-8')
        request.write(msg)

    def render(self, request: server.Request) -> int | None:
        consumer: Consumer | None = None
        command = b"listen"
        path: list[bytes] | None = request.postpath
        assert path is not None

        if path and path[-1] == b'':
            path = path[:-1]

        if path and path[0] in (b"listen", b"add", b"remove"):
            command = path[0]
            path = path[1:]

        if command == b"listen":
            cid = unicode2bytes(str(uuid.uuid4()))
            consumer = Consumer(request)

        elif command in (b"add", b"remove"):
            if path:
                cid = path[0]
                path = path[1:]
                if cid not in self.consumers:
                    self.finish(request, 400, b"unknown uuid")
                    return None
                consumer = self.consumers[cid]
            else:
                self.finish(request, 400, b"need uuid")
                return None

        assert consumer is not None

        pathref = b"/".join(path)
        decoded_path = self.decodePath(path)

        if command == b"add" or (command == b"listen" and decoded_path):
            options = request.args
            assert options is not None
            for k in options:
                if len(options[k]) == 1:
                    options[k] = options[k][1]

            try:
                d = self.master.mq.startConsuming(
                    consumer.onMessage, tuple(bytes2unicode(p) for p in decoded_path)
                )

                @d.addCallback
                def register(qref: QueueRef) -> None:
                    consumer.registerQref(pathref, qref)

                d.addErrback(log.err, "while calling startConsuming")
            except NotImplementedError:
                self.finish(request, 404, b"not implemented")
                return None
            except InvalidPathError:
                self.finish(request, 404, b"not implemented")
                return None
        elif command == b"remove":
            try:
                consumer.stopConsuming(pathref)
            except KeyError:
                self.finish(request, 404, b"consumer is not listening to this event")
                return None

        if command == b"listen":
            self.consumers[cid] = consumer
            request.setHeader(b"content-type", b"text/event-stream")
            request.write(b"")
            request.write(b"event: handshake\n")
            request.write(b"data: " + cid + b"\n")
            request.write(b"\n")
            d = request.notifyFinish()

            @d.addBoth
            def onEndRequest(_: Any) -> None:
                consumer.stopConsuming()
                del self.consumers[cid]

            return server.NOT_DONE_YET

        self.finish(request, 200, b"ok")
        return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import log

from buildbot.process.buildstep import BuildStep
from buildbot.process.buildstep import ShellMixin
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS


class Robocopy(ShellMixin, BuildStep):
    """Robocopy build step.

    This is just a wrapper around the standard shell command that
    will handle arguments and return codes accordingly for Robocopy.
    """

    renderables = ['custom_opts', 'destination', 'exclude_dirs', 'exclude_files', 'files', 'source']

    # Robocopy exit flags (they are combined to make up the exit code)
    # See: http://ss64.com/nt/robocopy-exit.html
    return_flags = {FAILURE: [8, 16], WARNINGS: [2, 4], SUCCESS: [0, 1]}

    def __init__(self, source, destination, exclude=None, exclude_files=None, **kwargs):
        self.source = source
        self.destination = destination

        self.files = kwargs.pop('files', None)
        self.recursive = kwargs.pop('recursive', False)
        self.mirror = kwargs.pop('mirror', False)
        self.move = kwargs.pop('move', False)

        self.exclude_files = exclude_files
        if exclude and not exclude_files:
            self.exclude_files = exclude

        self.exclude_dirs = kwargs.pop('exclude_dirs', None)
        self.custom_opts = kwargs.pop('custom_opts', None)
        self.verbose = kwargs.pop('verbose', False)

        super().__init__(**kwargs)

    @defer.inlineCallbacks
    def run(self):
        command = ['robocopy', self.source, self.destination]
        if self.files:
            command += self.files
        if self.recursive:
            command.append('/E')
        if self.mirror:
            command.append('/MIR')
        if self.move:
            command.append('/MOVE')
        if self.exclude_files:
            command.append('/XF')
            command += self.exclude_files
        if self.exclude_dirs:
            command.append('/XD')
            command += self.exclude_dirs
        if self.verbose:
            command += ['/V', '/TS', '/FP']
        if self.custom_opts:
            command += self.custom_opts
        command += ['/TEE', '/NP']

        cmd = yield self.makeRemoteShellCommand(command=command)

        yield self.runCommand(cmd)

        # If we have a "clean" return code, it's good.
        # Otherwise, look for errors first, warnings second.
        if cmd.rc in (0, 1):
            return SUCCESS
        for result in [FAILURE, WARNINGS]:
            for flag in self.return_flags[result]:
                if (cmd.rc & flag) == flag:
                    return result

        log.msg(f"Unknown return code for Robocopy: {cmd.rc}")
        return EXCEPTION



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import masters
from buildbot.data import types
from buildbot.db.changesources import ChangeSourceAlreadyClaimedError

if TYPE_CHECKING:
    from buildbot.db.changesources import ChangeSourceModel
    from buildbot.util.twisted import InlineCallbacksType


@defer.inlineCallbacks
def _db2data(master, dbdict: ChangeSourceModel):
    dbmaster = None
    if dbdict.masterid is not None:
        dbmaster = yield master.data.get(('masters', dbdict.masterid))
    data = {
        'changesourceid': dbdict.id,
        'name': dbdict.name,
        'master': dbmaster,
    }
    return data


class ChangeSourceEndpoint(base.Endpoint):
    pathPatterns = [
        "/changesources/n:changesourceid",
        "/masters/n:masterid/changesources/n:changesourceid",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        dbdict = yield self.master.db.changesources.getChangeSource(kwargs['changesourceid'])
        if 'masterid' in kwargs:
            if dbdict.masterid != kwargs['masterid']:
                return None
        return (yield _db2data(self.master, dbdict)) if dbdict else None
        return (yield _db2data(self.master, dbdict)) if dbdict else None


class ChangeSourcesEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/changesources",
        "/masters/n:masterid/changesources",
    ]
    rootLinkName = 'changesources'

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        changesources = yield self.master.db.changesources.getChangeSources(
            masterid=kwargs.get('masterid')
        )
        csdicts = yield defer.DeferredList(
            [_db2data(self.master, cs) for cs in changesources],
            consumeErrors=True,
            fireOnOneErrback=True,
        )
        return [r for (s, r) in csdicts]


class ChangeSource(base.ResourceType):
    name = "changesource"
    plural = "changesources"
    endpoints = [ChangeSourceEndpoint, ChangeSourcesEndpoint]

    class EntityType(types.Entity):
        changesourceid = types.Integer()
        name = types.String()
        master = types.NoneOk(masters.Master.entityType)

    entityType = EntityType(name)

    @base.updateMethod
    def findChangeSourceId(self, name: str) -> defer.Deferred[int]:
        return self.master.db.changesources.findChangeSourceId(name)

    @base.updateMethod
    @defer.inlineCallbacks
    def trySetChangeSourceMaster(
        self, changesourceid: int, masterid: int
    ) -> InlineCallbacksType[bool]:
        # the db layer throws an exception if the claim fails; we translate
        # that to a straight true-false value. We could trap the exception
        # type, but that seems a bit too restrictive
        try:
            yield self.master.db.changesources.setChangeSourceMaster(changesourceid, masterid)
        except ChangeSourceAlreadyClaimedError:
            return False
        return True

    @defer.inlineCallbacks
    def _masterDeactivated(self, masterid):
        changesources = yield self.master.db.changesources.getChangeSources(masterid=masterid)
        for cs in changesources:
            yield self.master.db.changesources.setChangeSourceMaster(cs.id, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from random import randint

from twisted.internet import defer
from twisted.python import log

_poller_instances = None


class Poller:
    def __init__(self, fn, instance, reactor):
        self.fn = fn
        self.instance = instance

        self.running = False
        self.pending = False

        # Invariants:
        #   - If self._call is not None or self._currently_executing then it is guaranteed that
        #     self.pending and self._run_complete_deferreds will be handled at some point in the
        #     future.
        #   - If self._call is not None then _run will be executed at some point, but it's not being
        #     executed now.
        self._currently_executing = False
        self._call = None
        self._next_call_time = None  # valid when self._call is not None

        self._start_time = 0
        self._interval = 0
        self._random_delay_min = 0
        self._random_delay_max = 0
        self._run_complete_deferreds = []

        self._reactor = reactor

    @defer.inlineCallbacks
    def _run(self):
        self._call = None
        self._currently_executing = True

        try:
            yield self.fn(self.instance)
        except Exception as e:
            log.err(e, f'while executing {self.fn}')
        finally:
            self._currently_executing = False

        was_pending = self.pending
        self.pending = False

        if self.running:
            self._schedule(force_now=was_pending)

        while self._run_complete_deferreds:
            self._run_complete_deferreds.pop(0).callback(None)

    def _get_wait_time(self, curr_time, force_now=False, force_initial_now=False):
        if force_now:
            return 0

        extra_wait = randint(self._random_delay_min, self._random_delay_max)

        if force_initial_now or self._interval == 0:
            return extra_wait

        # note that differently from twisted.internet.task.LoopingCall, we don't care about
        # floating-point precision issues as we don't have the withCount feature.
        running_time = curr_time - self._start_time
        return self._interval - (running_time % self._interval) + extra_wait

    def _schedule(self, force_now=False, force_initial_now=False):
        curr_time = self._reactor.seconds()
        wait_time = self._get_wait_time(
            curr_time, force_now=force_now, force_initial_now=force_initial_now
        )
        next_call_time = curr_time + wait_time

        if self._call is not None:
            # Note that self._call can ever be moved to earlier time, so we can always cancel it.
            self._call.cancel()

        self._next_call_time = next_call_time
        self._call = self._reactor.callLater(wait_time, self._run)

    def __call__(self):
        if not self.running:
            return
        if self._currently_executing:
            self.pending = True
        else:
            self._schedule(force_now=True)

    def start(self, interval, now=False, random_delay_min=0, random_delay_max=0):
        assert not self.running
        self._interval = interval
        self._random_delay_min = random_delay_min
        self._random_delay_max = random_delay_max
        self._start_time = self._reactor.seconds()

        self.running = True
        self._schedule(force_initial_now=now)

    @defer.inlineCallbacks
    def stop(self):
        self.running = False
        if self._call is not None:
            self._call.cancel()
            self._call = None
        if self._currently_executing:
            d = defer.Deferred()
            self._run_complete_deferreds.append(d)
            yield d


class _Descriptor:
    def __init__(self, fn, attrName):
        self.fn = fn
        self.attrName = attrName

    def __get__(self, instance, cls):
        try:
            poller = getattr(instance, self.attrName)
        except AttributeError:
            poller = Poller(self.fn, instance, instance.master.reactor)
            setattr(instance, self.attrName, poller)
            # track instances when testing
            if _poller_instances is not None:
                _poller_instances.append((instance, self.attrName))
        return poller


def method(fn):
    stateName = "__poll_" + fn.__name__ + "__"
    return _Descriptor(fn, stateName)


def track_poll_methods():
    global _poller_instances
    _poller_instances = []


def reset_poll_methods():
    global _poller_instances
    for instance, attrname in _poller_instances:  # pylint: disable=not-an-iterable
        delattr(instance, attrname)
    _poller_instances = None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


class TupleMatchingMixin:
    # a bunch of tuple-matching tests that all call do_test_match
    # this is used to test this behavior in a few places

    def do_test_match(self, routingKey, shouldMatch, *tuples):
        raise NotImplementedError

    def test_simple_tuple_match(self):
        return self.do_test_match(('abc',), True, ('abc',))

    def test_simple_tuple_no_match(self):
        return self.do_test_match(('abc',), False, ('def',))

    def test_multiple_tuple_match(self):
        return self.do_test_match(('a', 'b', 'c'), True, ('a', 'b', 'c'))

    def test_multiple_tuple_match_tuple_prefix(self):
        return self.do_test_match(('a', 'b', 'c'), False, ('a', 'b'))

    def test_multiple_tuple_match_tuple_suffix(self):
        return self.do_test_match(('a', 'b', 'c'), False, ('b', 'c'))

    def test_multiple_tuple_match_rk_prefix(self):
        return self.do_test_match(('a', 'b'), False, ('a', 'b', 'c'))

    def test_multiple_tuple_match_rk_suffix(self):
        return self.do_test_match(('b', 'c'), False, ('a', 'b', 'c'))

    def test_None_match(self):
        return self.do_test_match(('a', 'b', 'c'), True, ('a', None, 'c'))

    def test_None_match_empty(self):
        return self.do_test_match(('a', '', 'c'), True, ('a', None, 'c'))

    def test_None_no_match(self):
        return self.do_test_match(('a', 'b', 'c'), False, ('a', None, 'x'))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot.util.twisted import async_to_deferred


class FakeConnection:
    is_fake_test_connection = True

    _waiting_for_interrupt = False

    def __init__(self, testcase, name, step, commands_numbers_to_interrupt):
        self.testcase = testcase
        self.name = name
        self.step = step
        self._commands_numbers_to_interrupt = commands_numbers_to_interrupt
        self._block_on_interrupt = False
        self._next_command_number = 0
        self._blocked_deferreds = []

    @async_to_deferred
    async def remoteStartCommand(
        self, remote_command, builder_name, command_id, command_name, args
    ) -> None:
        self._waiting_for_interrupt = False
        if self._next_command_number in self._commands_numbers_to_interrupt:
            self._waiting_for_interrupt = True

            await self.step.interrupt('interrupt reason')

            if self._waiting_for_interrupt:
                raise RuntimeError("Interrupted step, but command was not interrupted")

        self._next_command_number += 1
        await self.testcase._connection_remote_start_command(remote_command, self, builder_name)

        # running behaviors may still attempt interrupt the command
        if self._waiting_for_interrupt:
            raise RuntimeError("Interrupted step, but command was not interrupted")

    def remoteInterruptCommand(self, builder_name, command_id, why):
        if not self._waiting_for_interrupt:
            raise RuntimeError("Got interrupt, but FakeConnection was not expecting it")
        self._waiting_for_interrupt = False

        if self._block_on_interrupt:
            d = defer.Deferred()
            self._blocked_deferreds.append(d)
            return d
        else:
            return defer.succeed(None)

    def set_expect_interrupt(self):
        if self._waiting_for_interrupt:
            raise RuntimeError("Already expecting interrupt but got additional request")
        self._waiting_for_interrupt = True

    def set_block_on_interrupt(self):
        self._block_on_interrupt = True

    def unblock_waiters(self):
        for d in self._blocked_deferreds:
            d.callback(None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

if TYPE_CHECKING:
    from buildbot.db.projects import ProjectModel


def project_db_to_data(model: ProjectModel, active=None):
    return {
        "projectid": model.id,
        "name": model.name,
        "slug": model.slug,
        "description": model.description,
        "description_format": model.description_format,
        "description_html": model.description_html,
        "active": active,
    }


class ProjectEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/projects/n:projectid",
        "/projects/i:projectname",
    ]

    @defer.inlineCallbacks
    def get(self, result_spec, kwargs):
        projectid = yield self.get_project_id(kwargs)
        if projectid is None:
            return None

        dbdict = yield self.master.db.projects.get_project(projectid)
        if not dbdict:
            return None
        return project_db_to_data(dbdict)


class ProjectsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    rootLinkName = 'projects'
    pathPatterns = [
        "/projects",
    ]

    @defer.inlineCallbacks
    def get(self, result_spec, kwargs):
        active = result_spec.popBooleanFilter("active")
        if active is None:
            dbdicts = yield self.master.db.projects.get_projects()
        elif active:
            dbdicts = yield self.master.db.projects.get_active_projects()
        else:
            # This is not optimized case which is assumed to be infrequently required
            dbdicts_all = yield self.master.db.projects.get_projects()
            dbdicts_active = yield self.master.db.projects.get_active_projects()
            ids_active = set(dbdict.id for dbdict in dbdicts_active)
            dbdicts = [dbdict for dbdict in dbdicts_all if dbdict.id not in ids_active]

        return [project_db_to_data(dbdict, active=active) for dbdict in dbdicts]


class Project(base.ResourceType):
    name = "project"
    plural = "projects"
    endpoints = [ProjectEndpoint, ProjectsEndpoint]
    eventPathPatterns = [
        "/projects/:projectid",
    ]

    class EntityType(types.Entity):
        projectid = types.Integer()
        name = types.Identifier(70)
        slug = types.Identifier(70)
        active = types.NoneOk(types.Boolean())
        description = types.NoneOk(types.String())
        description_format = types.NoneOk(types.String())
        description_html = types.NoneOk(types.String())

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generate_event(self, _id, event):
        project = yield self.master.data.get(('projects', str(_id)))
        self.produceEvent(project, event)

    @base.updateMethod
    def find_project_id(self, name: str, auto_create: bool = True):
        return self.master.db.projects.find_project_id(name, auto_create)

    @base.updateMethod
    @defer.inlineCallbacks
    def update_project_info(
        self,
        projectid: int,
        slug: str,
        description: str | None,
        description_format: str | None,
        description_html: str | None,
    ):
        yield self.master.db.projects.update_project_info(
            projectid, slug, description, description_format, description_html
        )
        yield self.generate_event(projectid, "update")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import os
from binascii import hexlify
from hashlib import sha1
from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.python import log

from buildbot.util import bytes2unicode
from buildbot.util import unicode2bytes

if TYPE_CHECKING:
    from buildbot.db.users import UserModel

# TODO: fossil comes from a plugin. We should have an API that plugins could use to
# register allowed user types.
srcs = ['git', 'svn', 'hg', 'cvs', 'darcs', 'bzr', 'fossil']
salt_len = 8


def createUserObject(master, author, src=None):
    """
    Take a Change author and source and translate them into a User Object,
    storing the user in master.db, or returning None if the src is not
    specified.

    @param master: link to Buildmaster for database operations
    @type master: master.Buildmaster instance

    @param authors: Change author if string or Authz instance
    @type authors: string or www.authz instance

    @param src: source from which the User Object will be created
    @type src: string
    """

    if not src:
        log.msg("No vcs information found, unable to create User Object")
        return defer.succeed(None)

    if src in srcs:
        usdict = {"identifier": author, "attr_type": src, "attr_data": author}
    else:
        log.msg(f"Unrecognized source argument: {src}")
        return defer.succeed(None)

    return master.db.users.findUserByAttr(
        identifier=usdict['identifier'],
        attr_type=usdict['attr_type'],
        attr_data=usdict['attr_data'],
    )


def _extractContact(user: UserModel | None, contact_types, uid):
    if user is not None and user.attributes is not None:
        for type in contact_types:
            contact = user.attributes.get(type)
            if contact:
                break
    else:
        contact = None
    if contact is None:
        log.msg(
            format="Unable to find any of %(contact_types)r for uid: %(uid)r",
            contact_types=contact_types,
            uid=uid,
        )
    return contact


@defer.inlineCallbacks
def getUserContact(master, contact_types, uid):
    """
    This is a simple getter function that returns a user attribute
    that matches the contact_types argument, or returns None if no
    uid/match is found.

    @param master: BuildMaster used to query the database
    @type master: BuildMaster instance

    @param contact_types: list of contact attributes to look for in
                         in a given user, such as 'email' or 'nick'
    @type contact_types: list of strings

    @param uid: user that is searched for the contact_types match
    @type uid: integer

    @returns: string of contact information or None via deferred
    """
    user = yield master.db.users.getUser(uid)
    contact = _extractContact(user, contact_types, uid)
    return contact


def encrypt(passwd):
    """
    Encrypts the incoming password after adding some salt to store
    it in the database.

    @param passwd: password portion of user credentials
    @type passwd: string

    @returns: encrypted/salted string
    """
    m = sha1()
    salt = hexlify(os.urandom(salt_len))
    m.update(unicode2bytes(passwd) + salt)
    crypted = bytes2unicode(salt) + m.hexdigest()
    return crypted


def check_passwd(guess, passwd):
    """
    Tests to see if the guess, after salting and hashing, matches the
    passwd from the database.

    @param guess: incoming password trying to be used for authentication
    @param passwd: already encrypted password from the database

    @returns: boolean
    """
    m = sha1()
    salt = passwd[: salt_len * 2]  # salt_len * 2 due to encode('hex_codec')
    m.update(unicode2bytes(guess) + unicode2bytes(salt))
    crypted_guess = bytes2unicode(salt) + m.hexdigest()

    return crypted_guess == bytes2unicode(passwd)
